#ifndef _ACOUSTIC_MODEL_
#define _ACOUSTIC_MODEL_

/*!
 *
 * @brief 音響モデルのためのプログラムのヘッダ
 * @author Fujii Yasuhisa
 * @date 31/03/2009
 *
 */

#include <iostream>
#include <deque>
#include <vector>
#include <map>
#include <stack>

#include "hmm.hpp"
#include "my_templates.hpp"
#include "my_typedefs.hpp"

namespace slp {
namespace hcnf{
class HCNFModel;
}
namespace spojus {

class Options;

namespace acoustic_model {
static const std::string kBlankString = "";	///< 空文字列を返すための変数
}

class AcousticModel {
public:
	typedef deque<RnumContainer *> MeanContainer;
	typedef deque<RnumContainer *> InvCovContainer;
	typedef deque<Hmm *> HmmContainer;
	typedef deque<Gauss *> GaussContainer;
	typedef deque<Gmm *> GmmContainer;
	typedef deque<StreamGmm *> StreamGmmContainer;
	typedef deque<StreamGmm::WeightContainer *> StreamGmmWeightContainer;
	typedef deque<Hmm::State *> HmmStateContainer;
	typedef deque<Hmm::Transition *> HmmTransitionContainer;
	
public:
	AcousticModel() { Init(); }
	AcousticModel(Options *options) { options_ = options; Init(); }

	virtual ~AcousticModel(){
		ContainerFree<HmmContainer>(hmm_);
		ContainerFree<MeanContainer>(mean_);
		ContainerFree<InvCovContainer>(inv_cov_);
		ContainerFree<GaussContainer>(gauss_);
		ContainerFree<GmmContainer>(gmm_);
		ContainerFree<StreamGmmContainer>(stream_gmm_);
		ContainerFree<StreamGmmWeightContainer>(stream_gmm_weight_);
		ContainerFree<HmmStateContainer>(state_);
		ContainerFree<HmmTransitionContainer>(transition_);
#ifdef DEBUG_DESTRUCTOR
		cout << "AcousticModel was freed successfully."<< endl;
#endif
	};

	virtual void ReadModel(const std::string& am_model){};

private:
	void Init() {
		stream_num_ = 1;
		vec_size_ = 0;
		default_weight_.push_back(1.0);
		use_left_context_ = false;
		use_right_context_ = false;
	}

public:
	/*!
	 * @brief 開始記号に続く単語のコンテキスト番号を返す
	 * @retval short 開始記号に続く単語のコンテキスト番号
	 *
	 * ・CDの場合は予め決められたコンテキストの番号を返す（デフォルト:SIL）
	 * ・CIの場合は0を返す
	 */
	short GetLeftInitialIndex();

	short GetLeftContextIndex(const std::string &left) {
		miter_ = left_context_index_.find(left);
		if (miter_ == left_context_index_.end()) {
#if 0
			cout << "Warning: Couldn't find left context --> return 0" << left << endl;
			return -1;
#endif
			return GetLeftContextIndex("");
		}
		else {
			return miter_->second;
		}
	}

	short GetRightContextIndex(const std::string &right) {
		miter_ = right_context_index_.find(right);
		if (miter_ == right_context_index_.end()) {
#if 0
			cout << "Warning: Couldn't find right context --> return 0" << right << endl;
			return 0;
#endif
			return GetRightContextIndex("");
		}
		else {
			return miter_->second;
		}
	}

	/*!
	 * @brief 左コンテキストの数を返す
	 * @retval short 左コンテキストの数
	 * 
	 * ・CDの場合は左コンテキストの数を返す
	 * ・CIの場合は1を返す
	 */
	short GetLeftContextSize() { return left_context_index_.empty() ? 1 : left_context_index_.size(); }

	/*!
	 * @brief 右コンテキストの数を返す
	 * @retval short 右コンテキストの数
	 * 
	 * ・CDの場合は右コンテキストの数を返す
	 * ・CIの場合は1を返す
	 */
	short GetRightContextSize() { return right_context_index_.empty() ? 1 : right_context_index_.size(); }

	const std::string& GetLeftContextName(int index) { return left_context_name_[index]; }
	const std::string& GetRightContexName(int index) { return right_context_name_[index]; }

	std::string GetEndingContextName(const std::string &name);
	std::string GetHeadingContextName(const std::string &name);

	short GetEndingContextIndex(const std::string &model) { return GetLeftContextIndex(GetEndingContextName(model)); }

	/*!
	 * @brief HMMモデル名に対応したHmmクラスへのポインタを返す
	 * @param const string& model_name モデルの名前
	 * @retval Hmm* モデル名に対応するHmmクラスのポインタ
	 */
	Hmm *GetHmmByName(const string &model_name);

	/*!
	 * @brief 一度計算したStreamGMMの尤度を格納するスペースを初期化する
	 * @param int sample_size 確保するスペースの量
	 * @retval void
	 */
	void InitStreamGmm(int sample_size);

	/*!
	 * @brief 一度計算したGMMの尤度を格納するスペースを初期化する
	 * @param int sample_size 確保するスペースの量
	 * @retval void
	 */
	void InitGmm(int sample_size);

	/*!
	 * @brief 一度計算したGauss分布の尤度を格納するスペースを初期化する
	 * @param int sample_size 確保するスペースの量
	 * @retval void
	 */
	void InitGauss(int sample_size);

	/*!
	 * @brief Gauss, Gmm, StreamGmm全てについて一度計算した尤度を格納するスペースを初期化する
	 * @param int sample_size 確保するスペースの量
	 * @retval void
	 */
	void InitAll(int sample_size) {
		InitGauss(sample_size);
		InitGmm(sample_size);
		InitStreamGmm(sample_size);
	}

	/*!
	 * @brief 一度計算したStreamGMMの尤度を格納するスペースを初期化する
	 * @retval void
	 */
	void InitStreamGmm();

	/*!
	 * @brief 一度計算したGMMの尤度を格納するスペースを初期化する
	 * @retval void
	 */
	void InitGmm();

	/*!
	 * @brief 一度計算したGauss分布の尤度を格納するスペースを初期化する
	 * @retval void
	 */
	void InitGauss();

	/*!
	 * @brief Gauss, Gmm, StreamGmm全てについて一度計算した尤度を格納するスペースを初期化する
	 * @retval void
	 */
	void InitAll() {
		InitGauss();
		InitGmm();
		InitStreamGmm();
	}


	/*!
	 * @brief 与えられた特徴ベクトルに対する音響尤度を全てのモデルについて計算するデバッグ用関数
	 * @param Rnum *data 音響尤度を計算するベクトル
	 * @param int frame 音響尤度を計算するフレーム（いらないけど，Score関数を呼ぶため保存用）
	 * @retval void
	 */
	void CheckAcousticScore(Rnum *data, int frame);

	/*!
	 * @brief 先頭に格納されたGMMを取得する
	 * @retval Gmm * 先頭に格納されたGMM
	 */
	Gmm *GetFrontGmm() { return gmm_.front(); }

	const std::string& get_model_name_() { return model_name_; }

	/*!
	 * @brief HMMのリストを読み込む
	 * @param const std::string& hlist_file HMMのリストを定義したファイル
	 * @retval void
	 */
	void ReadHmmList(const std::string& hlist_file);

	/*!
	 * @brief 論理HMMを物理HMMに変換する
	 * @param const std::string& hlist_file HMMのリストを定義したファイル
	 * @retval void
	 */
	std::string GetPhysicalHmm(const std::string& logical_hmm);

	int HmmContextSize(const std::string& hmm) {
		if (hmm.empty()) { return -1; }
		int c(0);
		if (string::npos != hmm.find("+")) {
			++ c;
		}
		if (string::npos != hmm.find("-")) {
			++ c;
		}
		return c;
	}

	void RemoveLeftContext(string& model_name) {
		size_t p;
		if (std::string::npos != (p = model_name.find("-"))) {
			model_name = model_name.substr(p + 1);
		}
	}

	void RemoveRightContext(string& model_name) {
		size_t p;
		if (std::string::npos != (p = model_name.rfind("+"))) {
			model_name = model_name.substr(0, p);
		}
	}

#ifdef HCNF
	void AttachHCNFModel(hcnf::HCNFModel* hcnf_model);
#endif

public:
	bool use_left_context_;
	bool use_right_context_;

protected:
	/*!
	 * @brief コンテキスト番号からコンテキスト名を引けるようにインデックスを作成する
	 * @retval void
	 */
	void MakeIndex2ContextName();

	/*!
	 * @brief 与えられたモデル名からコンテキストを登録する
	 * @param const string& model
	 * @retval void
	 */
	void RegisterContext(const string& model);

	// HMMの名前とポインタの対応を覚える
	map<std::string,Hmm *> hmm_map_;
	HmmContainer hmm_;

	// 音響コンテキストの番号を覚える
	map<std::string,short>::iterator miter_;
	map<std::string,short> left_context_index_;
	map<std::string,short> right_context_index_;
	vector<std::string> left_context_name_;
	vector<std::string> right_context_name_;

	// 全てのデータのプール用変数を定義
	MeanContainer mean_;
	InvCovContainer inv_cov_;
	GaussContainer gauss_;
	GmmContainer gmm_;
	StreamGmmContainer stream_gmm_;
	StreamGmmWeightContainer stream_gmm_weight_;
	HmmStateContainer state_;
	HmmTransitionContainer transition_;

	short vec_size_;				///< 入力の次元
	short stream_num_;				///< streamの数
	vector<short> stream_dim_;			///< 各ストリームの次元数
	StreamGmm::WeightContainer default_weight_;	///< 各ストリームのデフォルト重み（全て1.0）

	std::string model_name_;				///< モデルのファイル名を覚えておくための変数
	std::map<std::string,std::string> logical_map_;		///< 物理HMMから論理HMMへのマッピング用

	Options *options_;
};


class SpojusAcousticModel : public AcousticModel {
public:
	SpojusAcousticModel(Options *options){ options_ = options; };
	SpojusAcousticModel(const std::string &am_model, Options *options){
		options_ = options;
		ReadModel(am_model);
	}

	virtual ~SpojusAcousticModel(){
#ifdef DEBUG_DESTRUCTOR
		cout << "SpojusAcousticModel was freed successfully."<< endl;
#endif
	};

	virtual void ReadModel(const std::string& am_model);

private:
	void ReadModelEach(const std::string &, Hmm *);
	void AllocateMemory(int omat_dimension, int mixture_dimension, const vector<short> &block_dimension);
	void ReadState(istream &, Hmm *);
	void ReadTransition(istream &, Hmm *);
	void ReadFeatureBlock(istream &in_fs, int omat_dimension, int mixture_dimension,
				       short block_dimension, int mean_base, int inv_cov_base, int gmm_index = 0);
	void RegisterGMM(int omat_dimension, int mixture_dimension, const vector<short> &block_dimension, Hmm *hmm);
};


class HtkAcousticModel : public AcousticModel {
public:
	HtkAcousticModel(Options *options){ options_ = options; };
	HtkAcousticModel(const std::string &am_model, Options *options){
		options_ = options;
		ReadModel(am_model);
	}

	virtual ~HtkAcousticModel(){
#ifdef DEBUG_DESTRUCTOR
		cout << "HtkAcousticModel was freed successfully."<< endl;
#endif
	};

	/*!
	 * @brief HTK形式のHMMを読み込みます
	 * @param const std::string& am_model
	 * @retval void
	 */
	void ReadModel(const std::string& am_model);

private:
	/*!
	 * @brief ストリームから通常の区切り文字に加えて">"で区切った場合で読み込む
	 * @param istream& in_fs 入力ストリーム
	 * @param std::string& term 入力した文字を格納する変数
	 * @param char mode 文字列の変換処理を行うかどうか 0:何もしない， 1：全て大文字にする，2:全て小文字にする
	 * @retval istream & 入力ストリームへの参照（ループの判定などに使う）
	 *
	 * ・ 読み込んだ文字列に">"が含まれる場合は，そこまでの文字列を読み込んだ結果とし，
	 *   それ以降の文字列はストリームに戻す
	 * ・ 読み込んだ文字列は全て小文字にする
	 */
	istream& ReadTerm(istream& in_fs, std::string& term, char mode=1);

	/*!
	 * @brief 連続する実数値を読み込む
	 * @param istream& in_fs ファイルストリーム
	 * @param RnumContainer *rnum_container 実数値コンテナへのポインタ
	 * @retval void
	 */
	void ReadRnum(istream& in_fs, RnumContainer *rnum_container);

	/*!
	 * @brief 連続する実数値を読み込む
	 * @param istream& in_fs ファイルストリーム
	 * @param RnumContainer *rnum_container 実数値コンテナへのポインタ
	 * @param short dim 実数値を読み込む数
	 * @retval void
	 */
	void ReadRnum(istream& in_fs, RnumContainer *rnum_container, short dim);

	/*!
	 * @brief HTK形式のHMMについて<VARIANCE>を読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param RnumContainer *inv_cov 共分散行列を読み込む領域へのポインタ
	 * @retval void
	 */
	void ReadVariance(istream& in_fs, RnumContainer *inv_cov);

	/*!
	 * @brief HTK形式のHMMについて<INVCOVAR>を読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param RnumContainer *inv_cov 共分散行列を読み込む領域へのポインタ
	 * @retval void
	 */
	void ReadInvCovar(istream& in_fs, RnumContainer *inv_cov);

        /*!
         * @brief HTK形式のHMMについて<BLOCK>を読み込む
         * @param istream& in_fs HMMファイルストリーム
         * @param RnumContainer *inv_cov 共分散行列を読み込む領域へのポインタ
         * @param vector<short>& block_dim ブロックの次元数を格納するコンテナ
         * @retval void
         */
        void ReadBlock(istream& in_fs, RnumContainer *inv_cov, vector<short>& block_dim);

	/*!
	 * @brief HTK形式のHMMについてstateの定義を読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param Hmm::State *state stateの定義を保存する領域へのポインタ
	 * @retval void
	 */
	void ReadState(istream& in_fs, Hmm::State *state);

	bool IsCovKind(const std::string& term) {
		if (term == "<DIAGC>" || term == "<INVDIAGC>" || term == "<FULLC>" || term == "<LLTC>" || term == "<XFORMC>") {
			return true;
		} else {
			return false;
		}
	}

	bool IsDurKind(const std::string& term) {
		if (term == "<NULLD>" || term == "<POISSOND>" || term == "<GAMMAD>" || term == "<GEND>") {
			return true;
		} else {
			return false;
		}
	}

	bool IsBaseKind(const std::string& term) {
		std::string::size_type pos;
		if (( pos = term.find("<DISCRETE")) != std::string::npos
		    || (pos = term.find("<LPC")) != std::string::npos
		    || (pos = term.find("<LPCCEPSTRA")) != std::string::npos
		    || (pos = term.find("<MFCC")) != std::string::npos
		    || (pos = term.find("<FBANK")) != std::string::npos
		    || (pos = term.find("<MELSPEC")) != std::string::npos
		    || (pos = term.find("<LPREFC")) != std::string::npos
		    || (pos = term.find("<LPDELCEP")) != std::string::npos
		    || (pos = term.find("<USER")) != std::string::npos)
		{
			return true;
		} else {
			return false;
		}
	}

	bool IsNextStream(const std::string& term) {
		if (term == "<STREAM>" || term == "<MIXTURE>" || term == "~m" 
		    || term == "<RCLASS>" || term == "~u" || term == "<MEAN>")
		{
			return true;
		} else {
			return false;
		}
	}

	bool IsNextMixture(const std::string& term) {
		if (term == "<MIXTURE>" || term == "~m" 
		    || term == "<RCLASS>" || term == "~u" || term == "<MEAN>")
		{
			return true;
		} else {
			return false;
		}
	}

	bool IsCov(const std::string& term) {
		if (term == "<VARIANCE>" || term == "<INVCOVAR>" || term == "<LLTCOVAR>"
                    || term == "<XFORM>" || term == "<BLOCK>") {
			return true;
		} else {
			return false;
		}
	}
	

	/*!
	 * @brief HTK形式のHMMを解析する
	 * @param istream& in_fs HMMファイルストリーム
	 * @retval void
	 */
	void ParseHmmdef(istream& in_fs);

	/*!
	 * @brief globalOptsについて解析する
	 * @param istream& in_fs HMMファイルストリーム
	 * @retval void
	 */
	void ParseGlobalOpts(istream& in_fs);
	
	/*!
	 * @brief regTree を解析する
	 * @param istream& in_fs ファイルストリーム
	 * @retval void
	 */
	void ParseRegTree(istream& in_fs);

	/*!
	 * @brief HTK形式のHMMについて<BeginHMM>~<EndHMM>の箇所を読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param Hmm *hmm HMMを格納する領域へのポインタ
	 * @retval void
	 */
	void ReadHmm(istream& in_fs, Hmm *hmm);

	/*!
	 * @brief HTK形式のHMMについてstateを読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param Hmm *hmm HMMを格納する領域へのポインタ
	 * @retval void
	 */
	void ParseState(istream& in_fs, Hmm *hmm);

	/*!
	 * @brief HTK形式のHMMについてweightsを読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param StreamGmm *stream_gmm weightを格納する領域を保持するstream_gmmへのポインタ
	 * @retval void
	 */
	void ParseWeights(istream& in_fs, StreamGmm *stream_gmm);

	/*!
	 * @brief HTK形式のHMMについてstreamを読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param StreamGmm *stream_gmm streamを格納する領域を保持するstream_gmmへのポインタ
	 * @retval void
	 */
	void ParseStream(istream& in_fs, StreamGmm *stream_gmm);

	/*!
	 * @brief HTK形式のHMMについてmixtureを読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param Gmm *gmm mixtureを格納する領域を保持するgmmへのポインタ
	 * @retval void
	 */
	void ParseMixture(istream& in_fs, Gmm *gmm);

	/*!
	 * @brief HTK形式のHMMについてmeanを読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @retval RnumContainer * 平均ベクトルへのポインタ
	 */
	RnumContainer *ParseMean(istream& in_fs);

	/*!
	 * @brief HTK形式のHMMについてcovを読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param RnumContainer *mean 平均ベクトルを格納する領域を保存するためのポインタ
	 * @retval Gauss * 平均ベクトルと共分散行列を格納したGaussへのポインタ
	 */
	Gauss *ParseCov(istream& in_fs, RnumContainer *mean);

	/*!
	 * @brief HTK形式のHMMについてtransPを読み込む
	 * @param istream& in_fs HMMファイルストリーム
	 * @param Hmm *hmm HMMを格納する領域へのポインタ
	 * @retval void
	 */
	void ParseTransP(istream& in_fs, Hmm *hmm);

	/*!
	 * @brief マクロ用の連想配列を開放する
	 * @retval void
	 */
	void ClearMacro();

private:
	map<std::string,Hmm::State *> state_macro_;
	map<std::string,Hmm::Transition *> transp_macro_;
	map<std::string,StreamGmm::WeightContainer *> weight_macro_;
	map<std::string,Gauss *> mixpdf_macro_;
	map<std::string,RnumContainer *> mean_macro_;
	map<std::string,RnumContainer *> cov_macro_;

	std::string hmmid_;			///< for <HmmSetID>
	std::string covkind_;		///< for covkind
	std::string basekind_;		///< for basekind
	std::string durkind_;		///< for durkind
	std::string regtree_;		///< for regtree
	short regtree_num_;		///< the number of non terminal nodes

	stack<std::string> term_stack_;
};

#ifdef HCNF
class HCNFAcousticModel : public AcousticModel {
public:
	HCNFAcousticModel(Options *options) : model_(0) { options_ = options; };
	HCNFAcousticModel(const std::string &hcnf_model, Options *options) : model_(0) {
		options_ = options;
		ReadModel(hcnf_model);
	}

	virtual ~HCNFAcousticModel(){
		Clean();
#ifdef DEBUG_DESTRUCTOR
		cout << "HCNFAcousticModel was freed successfully."<< endl;
#endif
	};

	/*!
	 * @brief HCNFを読み込みます
	 * @param const std::string& hcnf_model
	 * @retval void
	 */
	void ReadModel(const std::string& hcnf_model);

private:
	void Clean();

	hcnf::HCNFModel* model_;
};
#endif

}
}

#endif // #ifndef _ACOUSTIC_MODEL_
