#ifndef _ASR_DICTIONARY_
#define _ASR_DICTIONARY_ 1
/*!
 * @brief 音声認識用の辞書を構築するためのプログラム
 *
 * 1. Dictionary クラスに格納された辞書情報を元に，trie構造を持った音声
 *    認識用辞書を構築する
 * 2. Dictionary クラスに格納された辞書情報を元に，線形辞書を構築する
 * 
 * @author Fujii Yasuhisa
 * @date 05/01/2009
 */

#include <vector>
#include <list>
#include <deque>
#include <vector>
#include <cmath>

#include "my_templates.hpp"
#include "my_typedefs.hpp"
#include "dictionary.hpp"

namespace slp {
namespace hcnf {
class State;
class GObservation;
}
namespace spojus {

class Gmm;
class StreamGmm;
class Spojus;
class OnePassDP;
class AsrDicWord;
class Hmm;

struct BeamWord;
/*!
 * @brief スコアと接続元を保存するための構造体
 */
struct Cell {
	Cell() : score_(-INFINITY), prev_beam_word_(NULL) {}
	Cell(Rnum score, BeamWord *prev_beam_word) {
		score_ = score;
		prev_beam_word_ = prev_beam_word;
	}
	const Cell &Set(Rnum score, BeamWord *prev_beam_word) {
		score_ = score;
		prev_beam_word_ = prev_beam_word;
		return *this;
	}
	void Init() {
		score_ = -INFINITY;
		prev_beam_word_ = NULL;
	}
	Rnum score_;
	BeamWord *prev_beam_word_;
};
typedef vector<Cell> CellContainer;


/*!
 * @brief ビームに残った単語を格納するための構造体
 */
struct BeamWord {
	typedef vector<BeamWord *> BeamWordContainer;
	typedef vector<BeamWordContainer> BeamWordContainerContainer;
	typedef vector<BeamWordContainerContainer> BeamWordContainerContainerContainer;

	BeamWord() : asr_dic_word_(NULL), prev_beam_word_(NULL),
		     cp_(-1), score_(-INFINITY), ac_score_(-INFINITY), lm_score_(-INFINITY),
		     container_(NULL), onebest_(false), reachable_(false) {}

	AsrDicWord *asr_dic_word_;	///< 単語
	BeamWord *prev_beam_word_;	///< この単語に接続する単語（直前の単語）
	int cp_;			///< 現在のポイント
	Rnum score_;			///< スコア
	Rnum ac_score_;			///< 音響スコア
	Rnum lm_score_;			///< 言語スコア
	BeamWordContainer *container_;	///< この単語を格納するコンテナへのポインタ
	CellContainer cell_;
	bool onebest_;			///< 1Best解に含まれる単語かどうか
	bool reachable_;		///< 最終状態に至るかどうか

	struct Compare {
		bool operator() (BeamWord *bw1, BeamWord *bw2) const {
			return bw1->score_ > bw2->score_;
		}
	};
};


class AsrDicWord;
/*!
 * @brief 差表を格納するためのクラス
 */
class DiffIndex {
public:
	DiffIndex() {}
	DiffIndex(short size) {	cell_.reserve(size); }
	DiffIndex(DiffIndex &diff) { *this = diff; }

	void Clear() { cell_.clear(); }
	void Show();
	CellContainer cell_;
};


/*!
 * @brief 辞書から読み込んだ情報に加えて，音声認識特有の情報を保持するためのクラス
 * 
 * 言語モデルキャッシュ用の情報や，コンテキスト情報などを新たに保存出来るようにする
 */
class AsrDicWord {
public:
	typedef deque<AsrDicWord *> AsrDicWordContainer;
	typedef BeamWord::BeamWordContainer BeamWordContainer;

	AsrDicWord() { Clear(); }
	AsrDicWord(DicWord *dic_word) {
		dic_word_ = dic_word;
		Clear();
	}
	void Clear() {
		// この0を仮定している箇所が存在するので，これを変更する時は注意が必要
		left_context_ = right_context_ = heading_context_ = ending_context_ = 0;
		data_.clear();
		Init();
	}
	void Init() {
		BeamWordContainer::iterator bwc_iter(prev_beam_word_.begin());
		for (; bwc_iter != prev_beam_word_.end(); ++ bwc_iter) {
			*bwc_iter = NULL;
		}
		RnumContainer::iterator rc_iter(score_.begin());
		for (; rc_iter != score_.end(); ++ rc_iter) {
			*rc_iter = -INFINITY;
		}
	}
	void Init(short l_size) {
		prev_beam_word_.resize(l_size);
		score_.resize(l_size);
		Init();
	}

public:
	DicWord *dic_word_;			///< 単語へのポインタ
	short left_context_;			///< 左コンテキスト
	short right_context_;			///< 右コンテキスト
	short heading_context_;			///< 単語始端コンテキスト
	short ending_context_;			///< 単語終端コンテキスト
	BeamWordContainer prev_beam_word_;	///< 直前にこの単語に接続した単語を記憶しておく
	RnumContainer score_;			///< 汎用記憶領域（e.g. キャッシュ）
	vector<int> data_;			///< 汎用記憶領域（e.g. キャッシュ）
};


class AsrDictionary {
public:
	typedef AsrDicWord::AsrDicWordContainer AsrDicWordContainer;

	class NodePointer;
	
	/*!
	 * @brief 辞書中のノードを表現するためのクラス
	 */
	class Node {
	public:
		typedef short NodeType;
		static const NodeType kNormal = 0x0;
		static const NodeType kNull = 0x1;
		static const NodeType kDiffIndex = 0x2;
		static const NodeType kWord = 0x4;
		static const NodeType kRoot = 0x8;
		static const NodeType kLeaf = 0x10;
		static const NodeType kLinked = 0x20;
		static const NodeType kLinearHead = 0x40;

		Node() { 
			node_ptr_ = NULL;
			stream_gmm_ = NULL;
			asr_dic_word_ = NULL;
			score_ = 0.0;
			type_ = kNormal;
			expand_ = false;
#ifdef HCNF
			gobs_ = NULL;
#endif
		}

		Node(Node &node) { *this = node; }

		virtual ~Node() {};
#if 0 // コンテナで管理するようにしたので
		{
			if (asr_dic_word_ != NULL) {
				delete asr_dic_word_;
				asr_dic_word_ = NULL;
			}
		}
#endif

		struct Transition {
			enum Type {kNormal, kNull};
			Transition() : to_(NULL), score_(0.0), type_(kNormal) {}
			Transition(Node *to, Rnum score, Type type) {
				to_ = to;
				score_ = score;
				type_ = type;
			}
			Transition& Set(Node *to, Rnum score, Type type) {
				to_ = to;
				score_ = score;
				type_ = type;
				return *this;
			}
			Node *to_;			///< 遷移先のポインタ
			Rnum score_;			///< 遷移のスコア
			Type type_;			///< 遷移のタイプ

			/// 比較オブジェクト
			struct Greater {
				bool operator() (const Transition &t1, const Transition &t2) const {
					return t1.score_ > t2.score_;
				}
			};
		};

		typedef vector<Transition> TransitionContainer;
		TransitionContainer transition_;		///< このノードからでる遷移
		uint   id_;					///< このノードの識別子
		string name_;					///< このノードの名前
		Rnum score_;					///< このノードのスコア
		StreamGmm *stream_gmm_;				///< このノードに結びつけられたGMMへのポインタ
		AsrDicWord *asr_dic_word_;			///< このノードに結びつけられた単語へのポインタ（ノードが葉の場合）
		NodePointer *node_ptr_;				///< このノードを指しているNodePointerクラスへのポインタ
		NodeType type_;					///< このノードのタイプ
		bool expand_;					///< 状態に拡張されたかどうかを示すフラグ
#ifdef HCNF
		hcnf::GObservation* gobs_;
#endif		
	};

	/*!
	 * @brief 各Nodeクラスの状態を保存するためのクラス
	 */
	class NodePointer {
	public:
		NodePointer() : node_(NULL) { Init(); }
		NodePointer(Node *node) { Init(node); }
		NodePointer(Node *node, short l_size) {
			DiffAlocate(l_size);
			Init(node);
		}
		~NodePointer() {
			if (diff_[0] != NULL) {	delete diff_[0]; }
			if (diff_[1] != NULL) { delete diff_[1]; }
		}

	private:
		/*!
		 * @brief 差表用の領域を確保する
		 */
		void DiffAlocate(short l_size) {
			diff_[0] = new DiffIndex(l_size);
			diff_[1] = new DiffIndex(l_size);
		}

	public:
		void Init() {
			cell_[0].Init();
			cell_[1].Init();
			if (diff_[0] != NULL) { diff_[0]->Clear(); }
			if (diff_[1] != NULL) { diff_[1]->Clear(); }
		}
		void Init(Node *node) {
			Init();
			node_ = node;
		}

		/// 比較オブジェクト
		struct Greater0 {
			bool operator() (NodePointer *n1, NodePointer *n2) const {
				return n1->cell_[0].score_ > n2->cell_[0].score_;
			}
		};

		/// 比較オブジェクト
		struct Greater1 {
			bool operator() (NodePointer *n1, NodePointer *n2) const {
				return n1->cell_[1].score_ > n2->cell_[1].score_;
			}
		};
#if 1
		struct WordOrder {
			bool operator() (NodePointer *n1, NodePointer *n2) const {
#if 0
				if (n1->node_->asr_dic_word_->ending_context_
				    == n2->node_->asr_dic_word_->ending_context_)
				{
					if (n1->node_->asr_dic_word_->dic_word_->number_
					    == n2->node_->asr_dic_word_->dic_word_->number_)
					{
						return n1->cell_[1].score_ > n2->cell_[1].score_;
					} else {
						return n1->node_->asr_dic_word_->dic_word_->number_
							< n2->node_->asr_dic_word_->dic_word_->number_;
					}
				} else {
					return n1->node_->asr_dic_word_->ending_context_
						< n2->node_->asr_dic_word_->ending_context_;
				}
#else
				if (n1->node_->asr_dic_word_->dic_word_->number_
				    == n2->node_->asr_dic_word_->dic_word_->number_)
				{
					return n1->cell_[1].score_ > n2->cell_[1].score_;
				} else {
					return n1->node_->asr_dic_word_->dic_word_->number_
						< n2->node_->asr_dic_word_->dic_word_->number_;
				}
#endif
			}
		};
		struct WordNumberEqual {
			bool operator() (NodePointer *n1, NodePointer *n2) const {
#if 0
				return (n1->node_->asr_dic_word_->ending_context_
					== n2->node_->asr_dic_word_->ending_context_
					&& n1->node_->asr_dic_word_->dic_word_->number_
					== n2->node_->asr_dic_word_->dic_word_->number_);
#else
				return (n1->node_->asr_dic_word_->dic_word_->number_
					== n2->node_->asr_dic_word_->dic_word_->number_);
#endif
			}
		};
#endif
		Node *node_;
		Cell cell_[2];
		DiffIndex *diff_[2];	///< 差表
	};

	typedef deque<Node *> NodeContainer;
	typedef vector<Node *> NodeVectorContainer;
	typedef Dictionary::DicWordContainer DicWordContainer;
	typedef Node::Transition Transition;
	typedef Node::TransitionContainer TransitionContainer;


	AsrDictionary() : spojus_(NULL), head_word_("<s>"), tail_word_("</s>"), final_word_node_(NULL) {}

	virtual ~AsrDictionary() {
		ContainerFree<NodeContainer>(node_);
		ContainerFree<NodeContainer>(special_node_);
		ContainerFree<AsrDicWordContainer>(asr_dic_word_);
#ifdef DEBUG_DESTRUCTOR
		cout << "AsrDictionary was freed successfully."<< endl;
#endif
	}

	virtual void Construct () {}
	virtual void Construct (Spojus *spojus) {}
	virtual void Show(){};

	/*!
	 * @brief 与えられた辞書の単語ノードからこの辞書の単語ノードへのリンクを作成する
	 * @param AsrDictionary *asr_dic リンクを作成する辞書
	 * @retval void
	 */
	void MakeWordLink(AsrDictionary *asr_dic);

	size_t GetContextSize() {
		return context2index_.size();
	}

	size_t GetLeftContextSize() {
		return prefix_m_.size();
	}

protected:
	virtual void Init();

	/*!
	 * @brief 左コンテキストを拡張する
	 * @retval void
	 */
	void ExpandLeftContext();

	/*!
	 * @brief 右コンテキストを拡張する
	 * @retval void
	 */
	void ExpandRightContext();

	/*!
	 * @brief 論理HMMを物理HMMにマッピングする
	 * @retval void
	 */
	void HmmMappingLogicalToPhysical();

	/*!
	 * @brief 単語間で発生しうるコンテキストを作成する
	 *
	 */
	void MakeInterWordContextIndex();

	/*!
	 * @brief コンテキスト依存モデルを使えるように辞書中の単語のモデル名を拡張する
	 * @retval void
	 */
	void ContextExpansion();

	/*!
	 * @brief 左コンテキスト処理のために，先頭ノードを分岐する
	 * @retval void
	 */
	void LeftContextExpansion();

	/*!
	 * @brief 先頭ノードをコンテキスト毎に分岐させる
	 * @param const NodeContainer& ncontainer
	 @ @retval void
	*/
	void BranchHeadNode(const NodeContainer& ncontainer);

	/*!
	 * @brief 差表用ノードを挿入する
	 * @param Node *node 差表用ノードを挿入したいノード
	 * @param uint node_num 作成するノードの番号
	 * @retval void
	 */
	void InsertDiffIndexNode(Node *node, uint node_num);

	/*!
	 * @brief 与えられた単語のコンテキスト情報を調べて記憶する
	 * @param AsrDicWord *asr_dic_word コンテキスト情報を調べる単語
	 * @retval void
	 */
	void RegisterContext(AsrDicWord *asr_dic_word);

	/*!
	 * @brief 開始記号と終了記号を表す単語を作成する
	 * @retval void
	 */
	void MakeSpecialWords();

	/*!
	 * @brief コンテキスト毎にルートノードを作成する
	 * @retval uint 作成したノードの数
	 */
	uint MakeRootNode();

public:
	/*!
	 * @brief ノードを格納しているコンテナへのリファレンスを返す
	 * @retval NodeContainer&
	 */
	NodeContainer& GetNodeContainer() { return node_; }

	/*!
	 * @brief 単語の自由領域を初期化する
	 * @param short l_size
	 * @retval void
	 */
	void AsrDicWordInit(short l_size) {
		NodeContainer::iterator nc_iter(word_node_.begin());
		for (; nc_iter != word_node_.end(); ++ nc_iter) {
			(*nc_iter)->asr_dic_word_->Init(l_size);
		}
		for (nc_iter = special_node_.begin(); nc_iter != special_node_.end(); ++ nc_iter) {
			(*nc_iter)->asr_dic_word_->Init(l_size);
		}
	}

	/*!
	 * @brief 指定された左コンテキストの初期ノードを返す
	 * @param short l_context 初期ノードを得たい左コンテキストのインデックス
	 * @retval Node * 指定された左コンテキストの初期ノード
	 */
	Node *GetInitialNode(short l_context) {	return initial_[l_context]; }

	/*!
	 * @brief 指定された左コンテキストの終端ノードを返す
	 * @param short l_context 初期ノードを得たい左コンテキストのインデックス
	 * @retval Node * 指定された左コンテキストの初期ノード
	 */
	Node *GetEndingNode(short l_context) {	return ending_[l_context]; }

private:
#if 0
	enum SpecialWord {kUnkWordIndex = 0, kDelWordIndex = 1, kInitialWordIndex = 2, kToEndWordIndex = 3};
#else
	enum SpecialWord {kUnkWordIndex = 0, kDelWordIndex = 1, kInitialWordIndex = 2};
#endif

public:
	const string& get_head_word_() { return head_word_; }
	const string& get_tail_word_() { return tail_word_; }
	void set_head_word_(const string& head_word) { head_word_ = head_word; }
	void set_tail_word_(const string& tail_word) { tail_word_ = tail_word; }

	/*!
	 * @brief 未知語ノードへのポインタを返す
	 * @retval Node * 未知語ノードへのポインタ
	 */
	Node *GetUnkWordNode() { return special_node_[kUnkWordIndex]; }
	/*!
	 * @brief 開始記号ノードへのポインタを返す
	 * @retval Node * 開始記号ノードへのポインタ
	 */
	Node *GetInitialWordNode() { return special_node_[kInitialWordIndex]; }
#if 0
	/*!
	 * @brief 終端単語接続用ノードへのポインタを返す
	 * @retval Node * 開始記号ノードへのポインタ
	 */
	Node *GetToEndWordNode() { return special_node_[kToEndWordIndex]; }
#endif
	/*!
	 * @brief 終了記号ノードへのポインタを返す
	 * @retval Node * 終了記号ノードへのポインタ
	 */
	Node *GetFinalWordNode() { return final_word_node_; }

protected:
	/*!
	 * @brief 新しいノードを作成する
	 * @param Node *node 新しいノードを接続する親ノードへのポインタ
	 * @param const string &node_name 新しいノードの名前
	 * @param uint node_num 新しいノードの番号
	 * @param Rnum trans_score 新しいノードへの遷移スコア
	 * @param DicWord *dic_word 新しいノードの単語
	 * @retval Node * 新しいノードへのポインタ
	 */
	Node *MakeNode(Node *node, const string &node_name, uint node_num, Rnum trans_score, DicWord *dic_word);

	/*!
	 * @brief 音節ノードからHMMの状態ノードへ変換する
	 * @retval void
	 */
	void SyllToState();

	/*!
	 * @brief 再帰的に，音節ノードを状態ノードへ変換する
	 * @param Node *node 展開するノード
	 * @retval void
	 */
	void ExpandSyllNode(Node *node);

	/*!
	 * @brief node_container内に含まれる遷移についてNull遷移を取り除く
	 * @retval void
	 */
	void RemoveNullTransition();

	/*!
	 * @brief node_container内に含まれる遷移についてNull遷移を取り除く
	 * @param NodeVectorContainer &node_container
	 * @retval void
	 */
	void RemoveNullTransition(NodeVectorContainer &node_container);

	/*!
	 * @brief node_containerの各ノードにGMMと遷移を割り付ける
	 * @param NodeVectorContainer &node_container
	 * @param Hmm *hmm 割り付けるGMMと遷移を格納したHMMへのポインタ
	 * @retval void
	 */
	void AttachHmm(NodeVectorContainer &node_container, Hmm *hmm);
	
#ifdef HCNF
	float GetHCNFTScore(hcnf::State* hcnf_state, int i, int j);
#endif

protected:
	Spojus *spojus_;
	NodeContainer node_;				///< Nodeへのポインタを格納
	NodeContainer word_node_;			///< 単語ノードへのポインタを格納（後の初期化の高速化のため）
	AsrDicWordContainer asr_dic_word_;		///< 単語ポインタをうまく開放するためのコンテナ
	string head_word_;				///< 先頭の単語（<s>）
	string tail_word_;				///< 終端の単語（</s>）
	Node *final_word_node_;				///< 終端単語ノード
	std::map<std::string,int> prefix_m_;		///< 単語の始端に現れうる左コンテキストの種類を保存
	std::map<std::string,int> suffix_m_;		///< 単語の終端に現れうる左コンテキストの種類を保存
	std::map<std::string,int> context2index_;	///< コンテキストをインデックスに変換する

private:
	NodeContainer special_node_;			///< 開始記号および終了記号へのノードを格納

public:
	NodeContainer initial_;				///< Initialノードへのポインタを格納
	NodeContainer ending_;				///< endノードへのポインタを格納
	NodeContainer word2wordnode_;			///< 単語IDと単語ノードの関係を保存する
};


class TrieDictionary : public AsrDictionary {
public:
	TrieDictionary(){}

	TrieDictionary(Spojus *spojus) {
		spojus_ = spojus;
		Construct();
	}

	virtual ~TrieDictionary(){
#ifdef DEBUG_DESTRUCTOR
		cout << "TrieDictionary was freed successfully."<< endl;
#endif
	}

	// 辞書構築
	void Construct();
	void Construct(Spojus *spojus);
	void Show();

private:
	/*!
	 * @brief 単語のユ二グラム確率を計算する
	 * @retval Rnum 木辞書から切り離す単語の閾値
	 */
	Rnum ComputeUnigramProb();

	/*!
	 * @brief 発音辞書を元に音節単位の木辞書を作成する
	 * @param Rnum sp_thr 木辞書から切り離す単語の閾値
	 * @retval void
	 */
	void MakeSyllNetwork(Rnum sp_thr);
	
	/*!
	 * @brief 発音辞書を元に音節単位の木辞書を作成する
	 * @param Rnum sp_thr 木辞書から切り離す単語の閾値
	 * @param char mode 0 -> 切り離さない, 1 -> 切り離す
	 * @retval void
	 */
	void MakeSyllNetworkSub(Rnum sp_thr, char mode);

	/*!
	 * @brief ユ二グラムファクタリングを行う
	 * @retval void
	 */
	void UnigramFactoring();

	/*!
	 * @brief 引数で与えられたノードのユ二グラムファクタリングスコアを求める
	 * @param Node *node ユ二グラムファクタリングスコアを求めるノード
	 * @retval Rnum nodeのユ二グラムファクタリングスコア
	 */
	Rnum UnigramFactoringScore(Node *node);

};


class LinearDictionary : public AsrDictionary {
public:
	struct WordScore {
		Rnum score_;		///< バイグラム確率および滞留確率に基づく単語のスコア
		AsrDicWord *asr_dic_word_;	///< 単語のid（pron_number_，発音）
		WordScore() : score_(-INFINITY), asr_dic_word_(NULL) {}
		WordScore& Set(AsrDicWord *asr_dic_word, Rnum score) {
			asr_dic_word_ = asr_dic_word;
			score_ = score;
			return *this;
		}
	};
	typedef vector<WordScore> WordScoreContainer;
	/// 比較オブジェクト
	struct WordScoreCompare {
		bool operator() (const WordScore& ws1, const WordScore& ws2) const {
			return ws1.score_ > ws2.score_;
		}
	};

public:
	LinearDictionary(Spojus *spojus) { Construct(spojus); }

	virtual ~LinearDictionary(){
#ifdef DEBUG_DESTRUCTOR
		cout << "LinearDictionary was freed successfully."<< endl;
#endif
	}

	void Show();
	// 辞書構築
	void Construct();
	void Construct(Spojus *spojus);

	/*!
	 * @brief 単語から接続されるノードへの遷移を持つノードを返す
	 * @param Word word_id ノードを取得したい単語ID
	 * @retval void
	 */
	Node *GetHeadNode(Word word_id) { return word2node_[word_id]; }

	/*!
	 * @brief 与えられた単語を履歴としてバイグラム確率をWordScoreContainerに適用する
	 * @param AsrDicWord *asr_dic_word バイグラムの履歴単語
	 * @param Rnum score バイグラム確率を適用する履歴のスコア
	 * @retval void
	 */
	void AddBigram(AsrDicWord *asr_dic_word, Rnum score);

	/*!
	 * @brief word_score_を初期化する
	 * @retval void
	 */
	void WordScoreInit();

	/*!
	 * @brief word_score_に単語とスコアを入力する
	 * @retval void
	 */
	void InputWordScore(AsrDicWord *asr_dic_word, Rnum score)
	{
		if (word_score_[asr_dic_word->dic_word_->pron_number_].score_ < score) 
		{
			word_score_[asr_dic_word->dic_word_->pron_number_].Set(asr_dic_word, score);
#if 0
			cout << asr_dic_word->dic_word_->name_ 
			     << " -> " << score << endl;
#endif
		}
	}

	/*!
	 * @brief 既に展開されている単語とバイグラム確率が存在する単語についてスコアが上位の単語を求める
	 * @retval void
	 */
	void WordScoreSort();

	/*!
	 * @brief word_score_に残った単語を基にInitialノードを変更する
	 * @retval void
	 */
	void InitialNodeUpdate();

private:
	/*!
	 * @brief 発音辞書を元に音節単位の線形辞書を作成する
	 * @retval void
	 */
	void MakeSyllNetwork();

	/*!
	 * @brief 動的に線形辞書を使用できる用に各単語に接続されるノードを計算する
	 * @retval void
	 */
	void MakeHeadNode();

	/*!
	 * @brief 先頭ノードから単語ノードを分離した後，単語IDとコンテキストで先頭ノードを引けるようにする
	 * @retval void
	 */
	void MakeHeadNodeIndex();

	/*!
	 * @brief 各単語についてバイグラム確率が存在する単語への接続を保存したノードを作成する
	 * @retval void
	 */
	void MakeHeadNodeBigram();

private:
	deque<NodeContainer> headnode_;
	NodeContainer word2node_;
	WordScoreContainer word_score_;
};


/*!
 * @brief 単語系列に対する辞書
 */
class WordDictionary : public AsrDictionary
{
public:
	WordDictionary (Spojus *spojus) { spojus_ = spojus; }
	WordDictionary (WordContainer &words, Spojus *spojus) {
		spojus_ = spojus;
		Construct(words);
	}

	~WordDictionary () {}

	void Construct (WordContainer &words);
	void Construct (Spojus *spojus) { spojus_ = spojus; }
	void Show();

private:
	/*!
	 * @brief 単語系列辞書の各種初期化処理を行う
	 * @retval void
	 */
	void Init();

	/*!
	 * @brief 発音辞書を元に音節単位の単語系列辞書を作成する
	 * @param WordContainer &words 辞書を作成する単語系列
	 * @retval void
	 */
	void MakeNetwork(WordContainer &words);

	/*!
	 * @brief 音節ノードを状態ノードへ変換する
	 * @param Node *node 展開するノード
	 * @retval Node * 最終ノードへのポインタ
	 */
	Node *ExpandSyllNode(Node *node);
};

}
}

#endif // #ifndef _ASR_DICTIONARY_
