#ifndef _ASR_LANG_MODEL_
#define _ASR_LANG_MODEL_ 1
/*!
 * @brief 音声認識用の言語モデルを提供するクラスのヘッダ
 *
 * LangModelクラスを複数保持し，内挿された言語モデル確率を計算する
 * 
 * @author Fujii Yasuhisa
 * @date 05/26/2009
 */

#include <vector>
#include <algorithm>

#include "my_typedefs.hpp"
#include "my_templates.hpp"
#include "lang_model.hpp"

namespace slp {
namespace spojus {

class AsrLangModel {
public:
	struct WeightedLangModel {
		WeightedLangModel() {
			lm_ = NULL;
			weight_ = 1.0;
		}
		~WeightedLangModel() {}
		LangModel *lm_;
		Rnum weight_;
	};
	/// 比較オブジェクト
	struct WeightedLangModelCompare {
		bool operator() (const WeightedLangModel& wlm1, const WeightedLangModel& wlm2) const {
			return wlm1.weight_ > wlm2.weight_;
		}
	};

	typedef vector<WeightedLangModel> WLangModelContainer;
	typedef LangModel::Ngram Ngram;

public:
	AsrLangModel(){
		max_ngram_ = max_ngram_save_ = 0;
	};
	AsrLangModel(Rnum lm_weight){
		lm_weight_ = lm_weight;
		max_ngram_ = max_ngram_save_ = 0;
	}
	~AsrLangModel(){
#ifdef DEBUG_DESTRUCTOR
		cout << "AsrLangModel was freed successfully."<< endl;
#endif
	};

	void AddLm(LangModel *lm, Rnum weight) {
		WeightedLangModel w_lm;
		w_lm.lm_ = lm;
		w_lm.weight_ = weight;
		w_lm_container_.push_back(w_lm);
		max_ngram_ = (max_ngram_ > lm->ngram_ ? max_ngram_ : lm->ngram_);
		max_ngram_save_ = max_ngram_;
		sort(w_lm_container_.begin(), w_lm_container_.end(), WeightedLangModelCompare());
	}

	Rnum Prob(Word *, ushort = 0);

	/*!
	 * @brief 混合重みが一番高い言語モデルにおけるコンテキストに続く単語を返す
	 * @param Word *words コンテキスト系列を格納した配列へのポインタ
	 * @param short N コンテキストの次数
	 * @param Ngram **ngram_ptr コンテキストに続く単語が開始する箇所へのポインタへのポインタ
	 * @param Word *len 単語が続く数
	 * @retval void
	 *
	 * 混合重みが一番高いモデルを探してコンテキストに続く単語を返します
	 * 混合重みが一番高いモデルが複数ある場合は，初めに読み込まれた方を優先する
	 *
	 */
	void GetWordsByContext(Word *words, short N, Ngram **ngram_ptr, Word *len);

	/*!
	 * @brief 尤度 を計算する
	 * @param WordContainer& words 単語系列を格納したコンテナ
	 * @param int start 実際に確率を計算する単語の開始点（開始記号などに対処するため）
	 * @retval Rnum 尤度
	 */
	Rnum Likelihood(WordContainer& words, int start);

	void set_lm_weight_(Rnum lm_weight) { lm_weight_ = lm_weight; }

	short get_max_ngram_() { return max_ngram_; }
	void set_max_ngram_(short max_ngram) {
		if (max_ngram == 0 || max_ngram > max_ngram_save_) {
			max_ngram_ = max_ngram_save_;
		} else {
			max_ngram_ = max_ngram;
		}
	}
	void recover_max_ngram_() { max_ngram_ = max_ngram_save_; }

private:

public:
	WLangModelContainer w_lm_container_;
	short max_ngram_;			///< 使用する言語モデルの最大次数
	short max_ngram_save_;			///< 含まれる言語モデル中で最大のN
	Rnum lm_weight_;			///< 言語重み
	Word words_[10];			///< 単語列保存用領域

private:
	
};

}
}

#endif // #ifndef _ASR_LANG_MODEL_
