#ifndef _CONF_NET_
#define _CONF_NET_

/**
 * @brief latticeからConfusion Networkを生成するためのヘッダファイル
 *                                 @author Fujii Yasuhisa
 *                                 @date 1/28/2009
 */

#include <deque>

#include "lattice.hpp"
#include "my_typedefs.hpp"

namespace slp {
namespace spojus {

class ConfNet {
public:
	typedef Lattice::Arc Arc;
	typedef Lattice::ArcContainer ArcContainer;
	typedef Lattice::Node Node;
	typedef Lattice::NodeContainer NodeContainer;
	typedef Lattice::ArcCompare ArcCompare;

	class EqClass {
	public:
		EqClass() : id_(0) {}
		virtual ~EqClass() {}
	public:
		ArcContainer arcs_;
		int id_;
                int time_;
	};
	typedef deque<EqClass> EqClassContainer;

	struct EqClassCompare {
		EqClassCompare(bool **matrix) {	class_adj_matrix_ = matrix; }
		bool operator() (const EqClass& eq_class1, const EqClass& eq_class2) {
			return class_adj_matrix_[eq_class1.id_][eq_class2.id_];
		}
		bool **class_adj_matrix_;
	};

public:
	ConfNet() {}
	ConfNet(Lattice *lattice) {
		Construct(lattice);
	}
	virtual ~ConfNet() {
		ContainerFree<ArcContainer>(arcs_);
	}

	void set_lattice_(Lattice *lattice) {
		lattice_ = lattice;
		set_spojus_(lattice->get_spojus_());
	}

	void set_spojus_(Spojus *spojus) {
		spojus_ = spojus;
	}

	void Search();

	void Construct(Lattice *lattice) {
		set_lattice_(lattice);
		Construct();
	}

	/*!
	 * @brief Latticeクラスのインスタンスからコンフュージョンネットワークを構築する
	 * @param Lattice& lattice ラティス
	 * @retval void
	 */
	void Construct();

	/*!
	 * @brief コンフュージョンネットワークを出力する
	 * @retval void
	 */
	void Show();

	/*!
	 * @brief コンフュージョンネットワークの最尤解を出力する
	 * @retval void
	 */
	void Output1Best();

	void ConnectOutputModule(int);

	void ArcPushBack(Arc *arc) {
		arcs_.push_back(arc);
	}

        /*!
         * @brief 事後確率を元に，各ビンの開始時間を設定する
         * @retval void
         */
        void SetBinTime();

	void Pruning();
	
private:
	/*!
	 * @brief ノードの隣接行列を作成する
	 * @retval void
	 */
	void MakeNodeAdjacentMatrix();

	/*!
	 * @brief 初期クラスを作成する
	 * @retval void
	 *
	 * L_{w,t1,t2} = {e \in E | Word(e) = w, Itime(e) = t1, Ftime(e) = t2} 
	 * の式に基づいてクラスを初期化する． 
	 *
	 */
	void MakeInitialEquivalentClasses();
	
	/*!
	 * @brief クラス間の隣接行列を作成する
	 * @retval void
	 */
	void MakeClassAdjacentMatrix();

	/*!
	 * @brief アーク間の順序を基にクラス間の順序を判定する
	 * @param EqClass& eq_class1 クラス1
	 * @param EqClass& eq_class2 クラス2
	 * @retval bool true or false : 順序がある or ない
	 */
	bool EqClassOrderCheck(EqClass& eq_class1, EqClass& eq_class2);

	/*!
	 * @brief 単語間のクラスタリングを行う
	 * @param int step 同一単語間のクラスタリングか異なる単語間のクラスタリングかを指定する 0 : 同一, 1 : 異なる
	 * @retval void
	 */
	void WordClustering(int step);

	/*!
	 * @brief 同一単語クラス間の類似度を計算する
	 * @param EqClass& eq_class1 類似度を計算するクラス1
	 * @param EqClass& eq_class2 類似度を計算するクラス2
	 * @retval Rnum 類似度
	 */
	Rnum IntraEqClassSimilarity(EqClass& eq_class1, EqClass& eq_class2);

	/*!
	 * @brief 異なる単語クラス間の類似度を計算する
	 * @param EqClass& eq_class1 類似度を計算するクラス1
	 * @param EqClass& eq_class2 類似度を計算するクラス2
	 * @retval Rnum 類似度
	 */
	Rnum InterEqClassSimilarity(EqClass& eq_class1, EqClass& eq_class2);

	/*!
	 * @brief 異なる単語クラス間の開始時刻と終了時刻が同じかどうかを判定する
	 * @param EqClass& eq_class1 類似度を計算するクラス1
	 * @param EqClass& eq_class2 類似度を計算するクラス2
	 * @retval Rnum 同じ:log(1.0), 違う:log(0.0)
	 */
	Rnum InterSameTimeEqClassSimilarity(EqClass& eq_class1, EqClass& eq_class2);

	/*!
	 * @brief 音素列の類似度を計算する
	 * @param const string& str1 類似度を計算する音素列1
	 * @param const string& str2 類似度を計算する音素列2
	 * @retval Rnum 類似度
	 */
	Rnum PhoneSimilarity(const string& str1, const string& str2);

	/*!
	 * @brief 同一単語間のクラスタリングでまとめられた単語を一つのアークにまとめる
	 * @retval void
	 */
	void MergeWord();

	/*!
	 * @brief クラスとクラス内の単語のソート
	 * @retval void
	 */
	void SortEqClasses();

private:
	Lattice *lattice_;
	Spojus *spojus_;
	ArcContainer arcs_;

public:
	EqClassContainer eq_class_;
};

}
}

#endif // #ifndef _CONF_NET_
