#ifndef _DICTIONARY_
#define _DICTIONARY_

/*!
 * 辞書関係のプログラムのヘッダファイル
 *		@author Fujii Yasuhisa
 *		@date 01/29/2009
 */

//#include <list>
#include <deque>
#include <string>
#include <map>
#include <set>
#include <climits>

#include "my_typedefs.hpp"

namespace slp {
namespace spojus {

/*!
 * 辞書中の1単語を表すクラス
 */
class DicWord {
public:
	enum Type {kNormal, kSkip};
	static const Word kEmpty = (sizeof(Word) == sizeof(uint) ? UINT_MAX : USHRT_MAX);		///< 空単語

	DicWord() :
		number_(0),
		pron_number_(0),
		class_id_(0),
		score_(0.0),
		class_score_(0.0),
		type_(kNormal) {}
	DicWord(const std::string &name, const std::string &kana, const std::string &syllable, std::deque<std::string> sylls) :
		number_(0), pron_number_(0),
		class_id_(0),
		score_(0.0),
		class_score_(0.0),
		type_(kNormal)
	{
		name_ = name;
		kana_ = kana;
		syllable_ = syllable;
		sylls_ = sylls;
	};
	virtual ~DicWord(){
#ifdef DEBUG_DESTRUCTOR
		std::cout << "DicWord was freed successfully."<< std::endl;
#endif
	}

	// アクセッサ
	const std::string &GetName(){return name_;};
	const std::string &getKana(){return kana_;};
	const std::string &get_syllable_(){return syllable_;};
	const std::deque<std::string> &get_sylls_(){return sylls_;};
	Word get_number_(){return number_;};
	Word get_pron_number_(){return pron_number_;};
	const std::string &getFeature() {return feature_;}
	const std::string &getSurface() {return surface_;}
        Type get_type_() { return type_; }

	void setName(const std::string &name){name_ = name;};
	void setKana(const std::string &kana){kana_ = kana;};
	void setSyllable(const std::string &syllable){syllable_ = syllable;};
	void setSylls(std::deque<std::string> &sylls){sylls_ = sylls;};
	void set_number_(Word number){number_ = number;};
	void set_pron_number_(Word pron_number){pron_number_ = pron_number;};
	void set_surface_(const std::string& surface) { surface_ = surface; }

	// 今の実装では3つ目が品詞の物しかとれないので，今後どうにかする必要がある．
	void setFeature(const std::string &term) {
		ulong pos = term.find("+");
		if (pos == std::string::npos) {
			feature_ = term;
		}
		else {
			if (surface_.empty()) {
				std::string surface = term.substr(0, pos);
				surface_ = surface;
#if 0
				std::cout << term << ","
					  << surface << std::endl;
#endif
			}
			pos = term.find("+", pos + 1);
			if (pos == std::string::npos) {
				feature_ = term;
			}
			else {
				feature_ = term.substr(pos + 1);
			}
		}
	}

public:
	Word number_;		///< 辞書での番号
	Word pron_number_;	///< 発音の種類まで考慮した番号
	Word class_id_;		///< クラスID
	std::string name_;
	std::string kana_;
	std::string syllable_;
	std::deque<std::string> sylls_;
	std::string surface_;
	std::string feature_;
	Rnum score_;		///< 汎用スコア記憶用領域（e.g. Unigram probability for factoring）
	Rnum class_score_;
	Type type_;		///< 単語に種類を持たせたい場合の領域
};


class Dictionary {
public:
	typedef std::map<std::string,Word> WordMap;
	typedef std::deque<DicWord> DicWordContainer;
public:
#if 0
	static const Word kSpecialWordNumber = 4;
#else
	static const Word kSpecialWordNumber = 2;
#endif

	Dictionary() : verbose_(1), class_num_(0), unk_id_(0) {
		Init();
	}

	virtual ~Dictionary() {
#ifdef DEBUG_DESTRUCTOR
		std::cout << "Dictionary was freed successfully."<< std::endl;
#endif
	}

	virtual void ReadModel(const std::string &) {};
	void Show();
	Word name2wordid(const std::string &);
	Word name2pronid(const std::string &name) {
		return WordId2PronId(name2wordid(name));
	}

	// 発音の種類を返す
	Word GetPronVariant(Word wordid) {
		if (wordid == DicWord::kEmpty) {
			return 0;
		} else if (wordid + 1 == number_) {
			// 最後の単語
			return pron_number_ - number2pron_[wordid];
		}
		return number2pron_[wordid + 1] - number2pron_[wordid];
	}

	Word WordId2PronId(Word wordid){
		if (wordid == DicWord::kEmpty) {
			return wordid;
		}
		return number2pron_[wordid];
	}

	Word PronId2WordId(Word pronid) {
		if (pronid == DicWord::kEmpty) {
			return pronid;
		}
		return dicwords_[pronid].get_number_();
	}

	const std::string &getFeature(Word wordid){
		if (wordid == DicWord::kEmpty) {
			return dicwords_[GetDelWordId()].feature_;
		}
		return dicwords_[number2pron_[wordid]].feature_;
	}

	const std::string &GetName(Word wordid){
		if (wordid == DicWord::kEmpty) {
			return dicwords_[GetDelWordId()].name_;
		}
		return dicwords_[number2pron_[wordid]].name_;
	}

	const std::string &GetPronName(Word wordid){
		if (wordid == DicWord::kEmpty) {
			return dicwords_[GetDelWordId()].name_;
		}
		return dicwords_[wordid].name_;
	}

	const std::string &getSurface(Word wordid){
		if (wordid == DicWord::kEmpty) {
			return dicwords_[GetDelWordId()].surface_;
		}
		return dicwords_[number2pron_[wordid]].surface_;
	}

	Word get_pron_number_() { return pron_number_; }
	Word get_number_() { return number_; }

	Word GetUnkWordId(){ return 0; }
	DicWord *GetUnkWord(){ return &dicwords_[0]; }
#if 0
	Word GetDelWordId(){ return 3; }
	DicWord *GetDelWord(){ return &dicwords_[3]; }
	Word GetInitialWordId(){ return 1; }
	Word GetFinalWordId(){ return 2; }
	DicWord *GetInitialWord(){ return &dicwords_[1]; }
	DicWord *GetFinalWord(){ return &dicwords_[2]; }
#else
	Word GetDelWordId(){ return 1; }
	DicWord *GetDelWord(){ return &dicwords_[1]; }
#endif

	DicWord *GetDicWordFromWordId (Word id) { 
		if (id == DicWord::kEmpty) {
			return GetDelWord(); 
		} else {
			return &dicwords_[number2pron_[id]]; 
		}
	}

	DicWord *GetDicWord(Word id) { return &dicwords_[id]; }

	DicWord *GetDicWordFromName (const std::string& name)
	{
		return GetDicWord(name2pronid(name));
	}

	const std::deque<std::string>& GetSyllsFromName (const std::string& name)
	{
		return dicwords_[name2pronid(name)].get_sylls_();
	}

        DicWord::Type GetType(Word wordid) {
                return dicwords_[number2pron_[wordid]].get_type_();
        }

	DicWordContainer::iterator NormalWordBeginIterator() {
		DicWordContainer::iterator dwc_iter = dicwords_.begin();
		for (Word i = 0; i < kSpecialWordNumber; i ++, dwc_iter ++) {}
		return dwc_iter;
	}

	/*!
	 * @brief コンテキストキューを設定する
	 * @param const std::string& ccs_file
	 * @retval void
	 */
	void SetContextCues(const std::string& ccs_file);

        /*!
         * @brief 与えられた単語列の未知語率を計算する（総数）
         * @param const WordContainer& wc 未知語率を計算する単語列
         * @retval void
         */
        double OOVRatioAll(const WordContainer& wc);

        /*!
         * @brief 与えられた単語列の未知語率を計算する（種類）
         * @param const WordContainer& wc 未知語率を計算する単語列
         * @retval void
         */
        double OOVRatioKind(const WordContainer& wc);

	/*!
	 * @brief クラス情報を読み込む
	 * @param const std::string& classfile クラスファイル
	 * @retval void
	 */
	void ReadClassInfo(const std::string& classfile);

	/*!
	 * @brief クラスIDを取得する．
	 * @param const string& cname
	 * @param char mode=0 モード 0 or 1 : 1の場合，存在しないクラス名ならば新しいクラスIDを付与する
	 * @retval Word クラスID
	 */
	Word GetClassID(const std::string& cname, char mode);
	Word GetClassID(const std::string& cname) {
		return GetClassID(cname, 0);
	}

	/*!
	 * @brief 単語IDをクラスIDに変換する
	 * @param Word cid 単語ID
	 * @retval Word クラスID
	 */
	Word Number2CID(Word word) {
		if (word == DicWord::kEmpty) { return DicWord::kEmpty; }
		else { return number2cid_[word]; }
	}

	void set_verbose_(char verbose) { verbose_ = verbose; }

protected:
	void Init()
	{
		int sp_word_num = RegisterSpecialWords();
		number_ = pron_number_ = sp_word_num;
		expand_ = false;
	}

	int RegisterSpecialWords()
	{
		Word num = 0;
		DicWord dicword;
		ModifyDicWord(&dicword, num, "<UNK>", "<UNK>");
		dicwords_.push_back(dicword);
		number2pron_.push_back(num);
		name2number_.insert(std::pair<std::string, Word>("<UNK>", num ++));

#if 0
		ModifyDicWord(&dicword, num, "<s>", "<s>");
		dicwords_.push_back(dicword);
		number2pron_.push_back(num);
		name2number_.insert(std::pair<std::string, Word>("<s>", num ++));

		ModifyDicWord(&dicword, num, "</s>", "</s>");
		dicwords_.push_back(dicword);
		number2pron_.push_back(num);
		name2number_.insert(std::pair<std::string, Word>("</s>", num ++));
#endif

		ModifyDicWord(&dicword, num, "<del>", "-");
		dicwords_.push_back(dicword);
		number2pron_.push_back(num);
		name2number_.insert(std::pair<std::string, Word>("<del>", num ++));

		return kSpecialWordNumber;
	}

	void ModifyDicWord(DicWord *dicword, Word word_number, const std::string &name, const std::string &feature)
	{
		dicword->number_ = word_number;
		dicword->pron_number_ = word_number;
		dicword->name_ = name;
		dicword->kana_ = name;
		dicword->surface_ = name;
		dicword->feature_ = feature;
	}

	/*!
	 * @brief クラスファイルから1行読み込む
	 * @param istream& is クラスファイルをオープンしたストリーム
	 * @retval void
	 */
	void ReadClassLine(std::istream& is);

public:
	char verbose_;
	DicWordContainer dicwords_;
	Word word_num_;
	Word number_;
	Word pron_number_;
	Word class_num_;
	Word unk_id_;
	WordContainer number2cid_;
	std::deque<Word> number2pron_;
	WordMap name2number_;
	WordMap cname2cid_;
	std::set<std::string> ccs_;
	bool expand_;				///< コンテキスト依存用のモデル拡張がなされたか
};

class SpojusDictionary : public Dictionary {
public:
	SpojusDictionary() {};
	SpojusDictionary(const std::string& dicfile, char verbose=1) {
		verbose_ = verbose;
		ReadModel(dicfile);
	}
	virtual ~SpojusDictionary() {};
	virtual void ReadModel(const std::string&);
private:
	void ParseLine(const std::string&);
	void ParseSubPart(const std::string& line, std::string& subpart, DicWord& dicword);
	void ParseInParentheses(const std::string& line, const std::string& subpart, DicWord& dicword);
};

class HtkDictionary : public Dictionary {
public:
	HtkDictionary() {};
	HtkDictionary(const std::string& dicfile, char verbose=1) {
		verbose_ = verbose;
		ReadModel(dicfile);
	}
	virtual ~HtkDictionary() {};
	virtual void ReadModel(const std::string&);
private:
	void ParseLine(const std::string&);
};

}
}

#endif // #ifndef _DICTIONARY_
