#ifndef _EDITING_
#define _EDITING_
/**
 * @brief Confusion Networkをから整形文を推定するためのプログラム
 *                                 @author Fujii Yasuhisa
 *                                 @date 10/02/2009
 */

#include <deque>
#include <list>
#include <map>

#include "my_typedefs.hpp"
#include "conf_net.hpp"

namespace slp {
namespace spojus {

class Editing {
public:
	typedef ConfNet::Arc Arc;
	typedef ConfNet::ArcContainer ArcContainer;
	typedef ConfNet::EqClass EqClass;
	typedef ConfNet::EqClassContainer EqClassContainer;

	Editing() : conf_net_(NULL), spojus_(NULL) {}
	
	Editing(ConfNet *conf_net, Spojus *spojus) {
		conf_net_ = conf_net;
		spojus_ = spojus;
	}
	
	virtual ~Editing(){}
	
	struct Hyp {
		Hyp() :
			score_(0.0),
			ac_score_(0.0),
			lm_score_(0.0) {}
		WordContainer words_;
		Rnum score_;
		Rnum ac_score_;
		Rnum lm_score_;
                vector<int> time_;
	};
	typedef deque<Hyp> HypContainer;

	struct HypCompare {
		bool operator() (const Hyp& hyp1, const Hyp& hyp2) const { 
			return hyp1.score_ > hyp2.score_;
		}
	};

	struct HypEqual {
		bool operator() (const Hyp& hyp1, const Hyp& hyp2) const {
			if (hyp1.words_.size() != hyp2.words_.size()) {
				return false;
			}
			for (int i = 0; i < (int) hyp1.words_.size(); ++ i) {
				if (hyp1.words_[i] != hyp2.words_[i]) {
					return false;
				}
			}
			return true;
		}
	};


	/*!
	 * @brief 整形処理を行う
	 * @retval void
	 */
	void Edit();

	/*!
	 * @brief 結果を出力する
	 * @param int nbest 何ベストまで出力するか
	 * @retval void
	 */
	void Show(int nbest=0);

	/*!
	 * @brief 結果を出力する
	 * @param HypContainer& nbest_stack 結果を出力するスタック
	 * @param int nbest 何ベストまで出力するか
	 * @retval void
	 */
	void Show(HypContainer& nbest_stack, int nbest=0);

private:
	/*!
	 * @brief 助詞の挿入を考慮出来るように脱落アークのみが存在するクラスをの全てのクラス間に挿入する
	 * @retval void
	 */
	void InsertDeletionArc();

	/*!
	 * @brief 禁則処理を行う
	 * @param WordContainer words
	 * @retval void
	 *
	 * 本来禁則処理は，言語モデルで可能であるが，修正が大変なので，このような方法で行う
	 *
	 */
	bool KinsokuJudgment(WordContainer& words);

	/*!
	 * @brief Confusion Networkから整形文を探索する
	 * @retval void
	 */
	void Search ();

	/*!
	 * @brief 話し言葉のCNから書き言葉のCNに変換する
	 * @retval void
	 */
	void Convert();

	/*!
	 * @brief 新しいアークを作成し，追加する
	 * @param EqClass& eq_class CNのビン
	 * @param Word word 新しいアークの単語
	 * @param Rnum score 新しいアークのスコア
	 * @retval void
	 */
	void AddArc(EqClass& eq_class, Word word, Rnum score);
	
	/*!
	 * @brief 仮説を展開する
	 * @param HypContainer& cur_stack 現在の仮説リストへの参照
	 * @param HypContainer& pre_stack 直前の仮説リストへの参照
	 * @param Arc *arc 新たに展開するアークへのポインタ
         * @param int time 展開する単語が所属するビンの開始時間
         * @retval void
	 */
	void ExpansionHyp(HypContainer& cur_stack, HypContainer& pre_stack, Arc *arc, int time);

	/*!
	 * @brief pre_hypにarcを接続して新しい仮説を作成する
	 * @param Hyp *new_hyp 新しい仮説へのポインタ
	 * @param Hyp &pre_hyp 直前の仮説
	 * @param Arc *arc 新たに展開するアークへのポインタ
         * @param int time 展開する単語が所属するビンの開始時間
	 * @retval void
	 */
	void CreateNewHyp(Hyp *new_hyp, Hyp& pre_hyp, Arc *arc, int time);

private:
	ConfNet *conf_net_;		// コンフュージョンネットワーク
	Spojus *spojus_;
	HypContainer nbest_stack_;
};

}
}

#endif // #ifndef _EDITING_
