#ifndef _FEATURE_INPUT_
#define _FEATURE_INPUT_
/*!
 * @brief 観測ベクトル入力用プログラムのヘッダ
 * @author Fujii Yasuhisa
 * @date 04/28/2009
 */

#include <iostream>
#include <list>

#include "my_typedefs.hpp"
#include "audio.hpp"
#include "network.hpp"

namespace slp {
namespace spojus {

class MfccExtractor;
class Options;

class InputData {
public:
	// The basic parameters for parmKind
	static const int kWAVEFORM = 0;
	static const int kLPC = 1;
	static const int kLPREFC = 2;
	static const int kLPCEPSTRA = 3;
	static const int kLPDELCEP = 4;
	static const int kIREFC = 5;
	static const int kMFCC = 6;
	static const int kFBANK = 7;
	static const int kMELSPEC = 8;
	static const int kUSER = 9;
	static const int kDISCRETE = 10;
	static const int kPLP = 11;

	// The qualifiers
	static const int k_E = 0000100;
	static const int k_N = 0000200;
	static const int k_D = 0000400;
	static const int k_A = 0001000;
	static const int k_C = 0002000;
	static const int k_Z = 0004000;
	static const int k_K = 0010000;
	static const int k_0 = 0020000;

public:
	InputData() :
		swap_(false), data_(NULL), A_(NULL), B_(NULL) {}
	InputData(bool swap) :
		data_(NULL), A_(NULL), B_(NULL) {swap_ = swap; }
	InputData(Rnum *data, bool swap) :
		A_(NULL), B_(NULL) { data_ = data; swap_ = swap; }
	InputData(const std::string& parm_kind_string) :
		data_(NULL), A_(NULL), B_(NULL) { set_parm_kind_string_(parm_kind_string); }

	virtual ~InputData(){
		Clear();
	};

	void Clear() {
		if (data_ != NULL) {
			delete [] data_;
			data_ = NULL;
		}
		if (A_ != NULL) {
			delete [] A_;
			A_ = NULL;
		}
		if (B_ != NULL) {
			delete [] B_;
			B_ = NULL;
		}
	}

	/*!
	 * @brief data の値を全て0で初期化する
	 * @retval void
	 */
	void DataZeroInit() {
		for (int i = 0; i < data_size_; ++ i) {
			data_[i] = 0.0;
		}
	}

	void Show() {
		std::cout << "Number of samples : " << n_samples_ << std::endl;
		std::cout << "Sample period in 100ns : " << samp_period_ << std::endl;
		std::cout << "Number of bytes per sample : " << samp_size_ << std::endl;
		std::cout << "A code indicating the sample kind : " << parm_kind_string_ << std::endl;
		std::cout << "Dimension number of input vector : " << dimension_ << std::endl;
		std::cout << "Data size : " << data_size_ << std::endl;
	}

	/*!
	 * @brief Rawパワーを取り除く
	 * @retval void
	 */
	void RemoveRawPower();

	/*!
	 * @brief ブロック型の音響モデルを使用する場合に特徴量の順番を入れ替える
	 * @retval void
	 */
	void BlockAmSwap();

	void ReadHtkMfc(const string &);

	Rnum *at(int frame) {
		return data_ + (frame * dimension_);
	}

	/*!
	 * @brief 受け取ったストリームにデータを吐き出す
	 * @param ostream *os データを吐き出すストリーム
	 * @retval void
	 */
	void Write(ostream *os);

	/*!
	 * @brief 特徴量を切り出す
	 * @param int start 切り出す開始点
	 * @param int end 切り出す終了点
	 * @retval void
	 */
	void Cut(int start, int end);

	/*!
	 * @brief 特徴量をSegment化する
	 * @param int left 左のフレーム幅
	 * @param int right 右のフレーム幅
	 * @retval void
	 */
	void Segment(int left, int right);

	void set_n_samples_(int n_samples) { n_samples_ = n_samples; }
	int get_n_samples_() { return n_samples_; }

	void set_samp_period_(int samp_period) { samp_period_ = samp_period; }
	int get_samp_period_() { return samp_period_; }

	void set_samp_size_(short samp_size) { samp_size_ = samp_size; }
	short get_samp_size_() { return samp_size_; }

	void set_parm_kind_(short parm_kind) {
		parm_kind_ = parm_kind;
		AnalysisParmKind();
	}
	short get_parm_kind_() { return parm_kind_; }

	void set_parm_kind_string_(const string& parm_kind_string) {
		parm_kind_string_ = parm_kind_string;
		parm_kind_ = AnalysisParmKindString();
	}
	const string& get_parm_kind_string_() { return parm_kind_string_; }

	void set_data_(Rnum *data) {
		if (data_ != NULL) {
			delete [] data_;
			data_ = NULL;
		}
		data_ = data;
	}
	Rnum *get_data_() { return data_; }

	void set_A_(float *A) {
		if (A_ != NULL) {
			delete [] A_;
			A_ = NULL;
		}
		A_ = A;
	}
	float *get_A_() { return A_; }

	void set_B_(float *B) {
		if (B_ != NULL) {
			delete [] B_;
			B_ = NULL;
		}
		B_ = B;
	}
	float *get_B_() { return B_; }

	void set_dimension_(int dimension) { dimension_ = dimension; }
	int get_dimension_() { return dimension_; }

	void set_data_size_(int data_size) { data_size_ = data_size; }
	int get_data_size_() { return data_size_; }

	void set_cepstrum_dim_(short cepstrum_dim) { cepstrum_dim_ = cepstrum_dim; }
	short get_cepstrum_dim_() { return cepstrum_dim_; }

	void AnalysisParmKind();
	short AnalysisParmKindString() {
		return AnalysisParmKindString(parm_kind_string_);
	}
	short AnalysisParmKindString(const string& parm_kind_string);

	/*!
	 * @brief parm_kind_とdimension_から生の特徴量の次元数を求める
	 * @retval void
	 */
	void ComputeCepstrumDim();

	/*!
	 * @brief data_の領域を2倍にする
	 * @retval void
	 */
	void DoubleDataSize();

	/*!
	 * @brief data_の領域を変更する
	 * @param int n_samples フレーム数
	 * @retval void
	 */
	void ResizeData(int n_samples);

private:
	void ComputeCompression();

public:
	int n_samples_;			///< number of samples in file (4-byte integer)
	int samp_period_;		///< sample period in 100ns (4-byte integer)
	short samp_size_;		///< number of bytes per sample (2-byte integer)
	short parm_kind_;		///< a code indicating the sample kind (2-byte integer)
	string parm_kind_string_;	///< parmKindの文字列表現 e.x) MFCC_E_Z

	short cepstrum_dim_;		///< 生の特徴量の次元数
	bool swap_;			///< 入力のswapを行うかどうか

	int dimension_;			///< 次元数
	int data_size_;			///< サンプル数 x 次元数
	Rnum *data_;
	float *A_;			///< compression parameter A x_short = A * x_float - B
	float *B_;			///< compression parameter B
};


#if 1

class FeatureExtraction {
public:
	FeatureExtraction() {};
	virtual ~FeatureExtraction() {};

	void Init();

	/*!
	 * @brief pcmデータを受け取って特徴抽出する
	 * @param short *pcm_data 特徴抽出するPCMデータ
	 * @param int size データサイズ
	 * @retval InputData* 特徴抽出結果を格納したInputDataへのポインタ
	 */
	virtual InputData *Extract(short *pcm_data, int size) { return NULL; };

	short get_delta_window_() { return delta_window_; }

	/*!
	 * @brief 与えられたフレームでのdeltaを計算する
	 * @param int frame deltaを計算するフレーム
	 * @param InputData *input_data deltaを計算するデータ列および計算結果の格納先
	 * @param int src_start deltaを計算する開始点 [src_start src_end)
	 * @param int src_end deltaを計算する終了点 
	 * @param int dst_start 計算結果格納開始点 [dst_start dst_start + (src_end - src_start))
	 * @retval void
	 */
	void ComputeDeltaPoint(int frame, InputData *input_data, int src_start, int src_end, int dst_start);

	/*!
	 * @brief deltaを計算する
	 * @param InputData *input_data deltaを計算するデータ列および計算結果の格納先
	 * @param int src_start deltaを計算する開始点 [src_start src_end)
	 * @param int src_end deltaを計算する終了点 
	 * @param int dst_start 計算結果格納開始点 [dst_start dst_start + (src_end - src_start))
	 * @retval void
	 */
	void ComputeDelta(InputData *input_data, int src_start, int src_end, int dst_start);

protected:
	short sampling_rate_;		///< サンプリング周波数
	short window_size_;		///< フレーム長
	short frame_shift_;		///< フレームシフト
	short delta_window_;		///< デルタを計算する窓長
};

class MfccExtractor : public FeatureExtraction {
public:
	struct FilterBank {
		int low_bin_;		///< 周波数領域下限カットオフ周波数 (bin) >=
		int high_bin_;		///< 周波数領域上限カットオフ周波数 (bin) <
		vector<double> center_;	///< 各フィルタバンクの中心周波数
		vector<int> index_;	///< 各周波数ビンに対応するフィルタバンクの次数
		vector<double> weight_;	///< 各周波数ビンの重み
	};

	MfccExtractor() {
		cos_mat_ = NULL;
		liftering_weight_ = NULL;
		Init();
	}

	MfccExtractor(const string& htk_config) {
		MfccExtractor();
		ParseHtkConfig(htk_config);
	}

	~MfccExtractor() { Clear(); }
	
	void Init();
	void Clear();

	void ParseHtkConfig(const string& htk_config);

	/*!
	 * @brief 1フレーム分のpcmデータを受け取ってMFCCを抽出する
	 * @param InputData *input_data MFCCを格納する領域へのポインタ
	 * @param int frame MFCCを格納するフレーム
	 * @param short *pcm_data 特徴抽出するPCMデータ
	 * @retval void
	 */
	void ExtractFrame(InputData *input_data, int frame, short *pcm_data);

	/*!
	 * @brief pcmデータを受け取って特徴抽出する
	 * @param short *pcm_data 特徴抽出するPCMデータ
	 * @param int size データサイズ
	 * @retval InputData* 特徴抽出結果を格納したInputDataへのポインタ
	 */
	virtual InputData *Extract(short *pcm_data, int size);

	short get_cepstrum_dim_() { return cepstrum_dim_; }
	void set_cepstrum_dim_(short cepstrum_dim) { cepstrum_dim_ = cepstrum_dim; }
	int get_dimension_() { return dimension_; }
	void set_dimension_(int dimension) { dimension_ = dimension; }
	void set_global_mean_(Rnum *global_mean) { global_mean_ = global_mean; }
	void set_parm_kind_(short parm_kind) { parm_kind_ = parm_kind; }

	/*!
	 * @brief 大域平均値を更新する
	 * @param int frame 大域平均値に適応する発話の総フレーム数
	 * @retval void
	 */
	void UpdateGlobalMean(int frame);

	/*!
	 * @brief InputDataについてヘッダの設定と領域の確保を行う
	 * @param InputData *input_data InputDataへのポインタ
	 * @param int n_samples サンプル数
	 * @retval void
	 */
	void SetInputData(InputData *input_data, int n_samples);

private:
	
	/*!
	 * @brief 線形周波数からメル周波数に変換する
	 * @retval T メル周波数
	 */
	template<class T>
	T Freq2Mel(T freq) { return (T) (1127.0 * log(1.0 + freq / 700)); }

	/*!
	 * @brief フィルタバンクを計算する
	 * @retval void
	 */
	void SetFilterBank();

	/*!
	 * @brief プリエンファシスする
	 * @param double *frame_data プリエンファシスするデータ列
	 * @param int size
	 * @retval void
	 */
	void PreEmphasis(double *frame_data, int size);

	/*!
	 * @brief FFTする
	 * @param double *frame_data FFTするデータ列
	 * @param double *real FFT結果の実部を格納する領域
	 * @param double *imag FFT結果の虚部を格納する領域
	 * @retval void
	 */
	void FFT(double *frame_data, double *real, double *imag);

	/*!
	 * @brief フィルタバンクを計算する
	 * @param double *real FFT結果の実部を格納する領域
	 * @param double *imag FFT結果の虚部を格納する領域
	 * @param double *filter_bank フィルタバンクの結果を格納する領域
	 * @retval void
	 */
	void MakeFilterBank(double *real, double *imag, double *filter_bank);

	/*!
	 * @brief DCTする
	 * @param double *filter_bank フィルタバンクの結果を格納する領域
	 * @param Rnum *mfcc DCTの結果を格納する領域
	 * @retval void
	 */
	void DCT(double *filter_bank, Rnum *mfcc);

	/*!
	 * @brief リフタリングする
	 * @param Rnum *mfcc リフタリングするデータ列へのポインタ
	 * @retval void
	 */
	void Liftering(Rnum *mfcc);

	/*!
	 * @brief input_dataのデータについて平均の正規化を行う
	 * @param InputData *input_data
	 * @retval void
	 */
	void MeanNormalization(InputData *input_data);

	/*!
	 * @brief input_dataのデータについてパワーの正規化を行う
	 * @param InputData *input_data
	 * @retval void
	 */
	void NormalizeLogEnergy(InputData *input_data);

	/*!
	 * @brief MFCCをMAP適応でCMNする
	 * @param Rnum *mfcc MFCC
	 * @param int frame 現在のフレーム数
	 * @retval void
	 */
	void MAPCMN(Rnum *mfcc, int frame);

private:
	bool zmean_source_;			///< DCの除去を行うかどうか
	bool enormalize_;			///< パワーの正規化を行うかどうか
	bool raw_energy_;			///< プリエンファシスや窓掛けの前にパワーを計算する

	double sil_floor_;			///< パワーの最低値を決めるデシベル値

	short filter_bank_num_;			///< フィルタバンクの数
	short fft_size_;			///< FFTの次元数
	short fft_bits_;			///< FFTのビット数
	short cepstrum_dim_;			///< ケプストラムの次数
	short liftering_dim_;			///< リフタリングの次数
	double pre_emphasis_;			///< プリエンファシスの係数
	double escale_;				///< パワーのスケーリング係数
	int dimension_;				///< 特徴量の次元数
	short parm_kind_;			///< a code indicating the sample kind (2-byte integer)

	double **cos_mat_;			///< DCT用領域
	double *liftering_weight_;		///< リフタリング用重み

	FilterBank filter_bank_;		///< フィルタバンク

	Rnum *global_mean_;			///< MAP適応時のMFCCの大域平均記憶用
	Rnum *local_mean_;			///< MAP適応時のMFCCの各発話の平均記憶用
	Rnum map_beta_;				///< MAP適応時の大域ベクトルの重み
	long all_frame_;			///< 総フレーム数
};
#endif


class FeatureInput {
public:
	enum RecState {kBegin=0x03, kSpeech=0x00, kEnd=0x01, kReject=0x04, kFinish=0x05, kSilence=0x02};

	FeatureInput() {
		Init();
	};
	FeatureInput(bool swap) {
		Init();
		swap_ = swap;
	}
	virtual ~FeatureInput(){
#ifdef DEBUG_DESTRUCTOR
		std::cout << "Try to free FeatureInput."<< std::endl;
#endif
		Clear();
#ifdef DEBUG_DESTRUCTOR
		std::cout << "FeatureInput was freed successfully."<< std::endl;
#endif
	}

	void Init();
	void Clear();

	/*!
	 * @brief ソケットをクライアント用に初期化する
	 * @param const string& host 接続するホスト名
	 * @param in_port_t port 接続するポート
	 * @retval int 初期化に成功したかどうか 0 > 失敗
	 */
	int InitClient(const string& host, in_port_t port);

        /*!
	 * @brief ソケットをクライアント用に初期化する
	 * @retval int 初期化に成功したかどうか 0 > 失敗
	 */
	int InitClient();

	/*!
	 * @brief ソケットをサーバ用に初期化する
	 * @param in_port_t port 接続するポート
	 * @retval void
	 */
	void InitServer(in_port_t port);

	/*!
	 * @brief socekt_の接続を確立する
	 * @retval void
	 */
	void Accept();

	/*!
	 * @brief InputDataのHeaderを送信する
	 * @retval void
	 */
	void SendHeader();

	/*!
	 * @brief InputDataのデータを1フレーム送信する
	 * @param int frame 送信するフレーム
	 * @param RecState frame_flag フレームの状態
	 * @retval void
	 */
	void SendFrame(int frame, RecState frame_flag);

	/*!
	 * @brief InputDataを一つ取り出して送信する
	 * @retval void
	 */
	void SendInputData();

	/*!
	 * @brief InputDataを受信してpushする
	 * @retval void
	 */
	void ReceiveInputData();

	/*!
	 * @brief Headerを受信する
	 * @retval void
	 */
	void ReceiveHeader();

	/*!
	 * @brief KaiLab Headerを受信する
	 * @param const string& param_kind_string
	 * @retval void
	 */
	void ReceiveKaiHeader(const string& param_kind_string);

	/*!
	 * @brief 状態フラグを受信する
	 * @retval RecState 受信したフレームの状態（音声:kSpeech，棄却:kReject，終端:kEnd，終了:kFinish）
	 */
	RecState ReceiveFlag();

	/*!
	 * @brief Kailab 状態フラグを受信する
	 * @retval RecState 受信したフレームの状態（音声:kSpeech，棄却:kReject，終端:kEnd，終了:kFinish）
	 */
	RecState ReceiveKaiFlag();

	/*!
	 * @brief データを1フレーム受信する
	 * @param int frame 受信するフレームの番号（0から）
	 * @retval void
	 */
	void ReceiveFrame(int frame);

	/*!
	 * @brief Kailab フロントエンドからデータを1フレーム受信する
	 * @param int frame 受信するフレームの番号（0から）
	 * @retval void
	 */
	void ReceiveKaiFrame(int frame);
	
	/*!
	 * @brief フレームタイプを元に記号を出力する
	 * @param RecState frame_flag フレームタイプ
	 * @retval void
	 */
	void OutputFrameType(RecState frame_flag);

	/*!
	 * @brief マイクから音声を読み込めるようにサウンドデバイスを初期化する
	 * @retval void
	 */
	void InitMic();

	/*!
	 * @brief マイクから音声を読み込んでpcm_data_に格納する
	 * @retval void
	 */
	void RecPcm();

	void ReadHtkMfcFileList(const string&);

	InputData* Back() { return input_data_.back(); }
	InputData* Deque() {
		InputData *input_data = input_data_.front();
		input_data_.pop_front();
		return input_data;
	}
	void Enque(InputData *input_data) {
		input_data_.push_back(input_data);
	}
	
	/*!
	 * @brief pcmファイルから音声データを読み込む
	 * @param const string& pcm_file
	 * @retval void
	 */
	void ReadPcm(const string& pcm_file);

	/*!
	 * @brief 与えられた1フレーム分のデータからMFCCを作成する
	 * @param MfccExtractor& mfc_ext MFCCを抽出するためのクラス
	 * @param int frame MFCCを抽出するフレーム
	 * @param short *pcm MFCCを抽出するデータ
	 * @retval void
	 */
	void ExtractMfccPoint(MfccExtractor& mfc_ext, int frame, short *pcm);

	/*!
	 * @brief pcm_data_のデータからMFCCを作成する
	 * @retval void
	 */
	void ExtractMfcc();

	/*!
	 * @brief マイクから音声を読み込んでMFCCを抽出し，サーバに送信する
	 * @retval void
	 */
	void MicFrontEnd();

	/*!
	 * @brief ファイルから読み込んだPCMデータからオンライン処理の手順と同様にしてMFCCを抽出し，サーバに送信する
	 * @param list<string>& file_list ファイルのリスト
	 * @retval void
	 */
	void PcmFileFrontEnd(list<string>& file_list);

	/*!
	 * @brief 指定されたフレームの特徴量の順番をブロック型に入れ替える
	 * @param int frame フレーム番号
	 * @retval void
	 */
	void BlockAmSwap(int frame);

        /*!
         * @brief 指定されたフレームのRawパワーを取り除く
         * @param int frame フレーム番号
         * @retval void
         */
        void RemoveRawPower(int frame);

        /*!
         * @brief クライアント機能を開始する
         * @retval void
         */
        void StartMicClient();

	/*!
	 * @brief 受け取ったデータをソケットから送信する
	 * @retval int
	 */
	void Write(const char *buff, int bsize)
	{
		socket_->write(buff, bsize);
	}

        void set_connect_host_(const string& connect_host) {
                connect_host_ = connect_host;
        }
        void set_connect_port_(in_port_t connect_port) {
                connect_port_ = connect_port;
        }

	void set_options_(Options *options) { options_ = options; }

private:
	/*!
	 * @brief フレームが音声区間候補かどうか判定する
	 * @retval bool 音声区間候補かどうか
	 */
	bool CheckSpeech(double energy, int z_count) {
#if 0
		if (energy > pow_threshold_
		    || (z_count > zero_cross_rate_ && energy > pow_threshold_ / 3.0))
		{
#else
		if (energy > pow_threshold_)
		{ 
#endif
			return true;
		} else {
			return false;
		}
	}

	/*!
	 * @brief データ系列の零交差数を計算する
	 * @param short *data
	 * @param int size
	 * @retval 零交差数
	 */
	int ComputeZeroCrossCount(short *data, int size);

public:
	MfccExtractor mfc_ext_;		///< Delta計算用

private:
	bool swap_;			///< 入力のswapを行うかどうか
	list<InputData *> input_data_;
	vector<short> pcm_data_;
	vector<short> tmp_data_;
	Mic mic_;

	MySocket *socket_;		///< InputDataをやり取りするためのソケット
	string connect_host_;		///< 接続するホスト
	in_port_t connect_port_;	///< 接続するポート

	short sampling_rate_;		///< サンプリング周波数
	short window_size_;		///< フレーム長
	short frame_shift_;		///< フレームシフト
	double pre_emphasis_;		///< プリエンファシスの係数
	short zero_cross_rate_;		///< 零交差のしきい値
	double pow_threshold_;		///< パワーのしきい値
	short start_duration_;		///< 音声開始点検出までの音声区間
	short end_duration_;		///< 音声終了点検出までの無音区間

	int max_samples_;		///< 最大サンプル数
	char seg_flag_;			///< 複数のセグメントを送るときに1が立つ（デフォルト:0）
	ulong tmstamp_;			///< 現在のタイムスタンプ(msec)

	short parm_kind_;		///< a code indicating the sample kind (2-byte integer)
	string parm_kind_string_;	///< parmKindの文字列表現 e.x) MFCC_E_Z
	string htk_config_;		///< HTK形式のコンフィグファイル
	Options *options_;		///< オプションクラス
};

}
}

#endif // #ifndef _FEATURE_INPUT_
