#ifndef _GMM_
#define _GMM_ 1

/*!
 * @brief GMMを扱うためのプログラムのためのヘッダファイル
 * @author Fujii Yasuhisa
 * @date 03/23/2009
 */

#include <iostream>
#include <cstdlib>
#include <cmath>
#include <vector>
#include "my_typedefs.hpp"
#include "my_math.hpp"
#include "utility.hpp"

namespace slp {
namespace spojus {

class Gauss {
public:
	Gauss() {}
	virtual ~Gauss() {}

public:
	void set_mean_(RnumContainer *mean) { mean_ = mean; mean_vec_ = &(mean_->at(0)); }
	void set_inv_cov_(RnumContainer *inv_cov) { inv_cov_ = inv_cov; inv_cov_vec_ = &(inv_cov_->at(0)); }
	void set_gconst_(Rnum gconst) { gconst_ = gconst; }
	void set_dimension_(short dimension) { dimension_ = dimension; }

	Rnum get_gconst_() { return gconst_; }

	virtual Rnum ComputeProb(Rnum *data) { return -INFINITY; }

	/*!
	 * @brief 共分散行列を元に定数項gconstを計算する
	 * @retval void
	 *
	 * gauss分布の定数項 : log(1/((2pi)^n |Sigma|))
	 */
	virtual void ComputeGconst(){}

	Rnum Score(Rnum *data, int frame) {
#ifdef SCORE_DEBUG
		cout << likelihood_.size() << endl;
		cout << "Gauss " << frame << " start" << endl;
#endif
		if (likelihood_[frame] > 0.0) {
			likelihood_[frame] = ComputeProb(data);
		}
#ifdef SCORE_DEBUG
		cout << "Gauss " << frame << " : " << likelihood_[frame] << endl;
#endif
		return likelihood_[frame];
	}

	Rnum Score(Rnum *data) {
		if (p_likelihood_ > 0.0) {
			p_likelihood_ = ComputeProb(data);
		}
		return p_likelihood_;
	}

	void Init(int sample_size) {
		likelihood_.resize(sample_size);
		for (int i = 0; i < sample_size; i ++) {
			likelihood_[i] = 1.0;		// 未使用領域は正の値に初期化
		}
	}

	void Init() { p_likelihood_ = 1.0; }

protected:
	Rnum gconst_;			///< gauss分布の定数項 ( log(1/sqrt((2pi)^n |Sigma|)) )
	short dimension_;
	RnumContainer mean_sub_vec_;	///< 入力から平均を引いたベクトル（作業領域）
	RnumContainer *mean_;		///< 平均ベクトル
	RnumContainer *inv_cov_;	///< 共分散行列の逆行列
	std::vector<Rnum> likelihood_;	///< 一度計算した値を保持しておくための領域
	Rnum p_likelihood_;		///< 今のフレームで一度計算した値を保持しておくための領域（全フレーム覚えておくとメモリ喰いすぎる．．）

	Rnum *mean_vec_;		///< 高速化のためのポインタ
	Rnum *inv_cov_vec_;		///< 高速化のためのポインタ
};


/*!
 * @brief 対角共分散のGauss分布用クラス
 */
class GaussDiag : public Gauss {
public:
	GaussDiag(RnumContainer *mean, RnumContainer *inv_cov) {
		set_mean_(mean);
		set_inv_cov_(inv_cov);
		mean_sub_vec_.resize(mean->size());
	}
public:
	Rnum ComputeProb(Rnum *data);
	void ComputeGconst();
};


/*!
 * @brief ブロック型共分散のGauss分布用クラス
 */
class GaussBlock : public Gauss {
public:
	GaussBlock() {}
        GaussBlock(RnumContainer *mean, RnumContainer *inv_cov, const std::vector<short> &block_dimension) {
                dimension_ = 0;
                for (int i = 0; i < (int) block_dimension.size(); ++ i) {
                        dimension_ += block_dimension[i];
                }
                set_mean_(mean);
                set_inv_cov_(inv_cov);
                block_dimension_ = block_dimension;
		mean_sub_vec_.resize(mean->size());
        }
	GaussBlock(RnumContainer *mean, RnumContainer *inv_cov, Rnum gconst, const std::vector<short> &block_dimension) {
		set_mean_(mean);
		set_inv_cov_(inv_cov);
		set_gconst_(gconst);
		block_dimension_ = block_dimension;
		mean_sub_vec_.resize(mean->size());
	}
	GaussBlock(short dimension, RnumContainer *mean, RnumContainer *inv_cov, const std::vector<short> &block_dimension) {
		dimension_ = dimension;
		set_mean_(mean);
		set_inv_cov_(inv_cov);
		block_dimension_ = block_dimension;
		mean_sub_vec_.resize(dimension);
	}
	GaussBlock(short dimension, RnumContainer *mean, RnumContainer *inv_cov, Rnum gconst, const std::vector<short> &block_dimension) {
		dimension_ = dimension;
		set_mean_(mean);
		set_inv_cov_(inv_cov);
		set_gconst_(gconst);
		block_dimension_ = block_dimension;
		mean_sub_vec_.resize(dimension);
	}
	virtual ~GaussBlock() {}

public:
	Rnum ComputeProb(Rnum *);
	void ComputeGconst();
private:
	std::vector<short> block_dimension_;	///< 各ブロックの次元数
};


/*!
 * @brief 全共分散のGauss分布用クラス
 */
class GaussFull : public Gauss {
public:
	GaussFull(){}
	GaussFull(RnumContainer *mean, RnumContainer *inv_cov) {
		set_mean_(mean);
		set_inv_cov_(inv_cov);
		mean_sub_vec_.resize(mean_->size());
	}
	virtual ~GaussFull(){}

public:
	Rnum ComputeProb(Rnum *);
	void ComputeGconst();
};


class Gmm {
public:
	typedef std::vector<Rnum> WeightContainer;
	typedef std::vector<Gauss *> GaussContainer;

	static const Rnum kScale = 30.0;

	Gmm(short mixture) {
		mixture_ = mixture; weight_.resize(mixture); gauss_.resize(mixture);
		for (short m = 0; m < mixture; ++ m) {
			weight_[m] = 1.0;
			gauss_[m] = NULL;
		}
	};

	Gmm(const GaussContainer& gauss, const WeightContainer& weight) {
		gauss_ = gauss;
		weight_ = weight;
	};

	virtual ~Gmm() {
#ifdef DEBUG_DESTRUCTOR
		cout << "Gmm was freed successfully."<< endl;
#endif
	};
	
public:
#if 1
	/*!
	 * @brief 引数の特徴ベクトルに対するGMMの尤度を計算する．
	 * @param Rnum *input_vec 尤度を計算する特徴量の系列へのポインタ
	 * @retval 入力の系列に対する対数尤度
	 */
	Rnum ComputeProb(Rnum *input_vec)
	{
#if 1
		Rnum prob = -INFINITY;
#else
		double prob = 0.0, max = -INFINITY;
		tmp_.resize(gauss_.size());
#endif

		for (uint m = 0; m < gauss_.size(); ++ m) {
#if 1
			prob = log_sum(prob, weight_[m] + gauss_[m]->ComputeProb(input_vec)); // log領域なので足し算
#else
			tmp_[m] = weight_[m] + gauss_[m]->ComputeProb(input_vec);
			if (max < tmp_[m]) {
				max = tmp_[m];
			}
#endif
#if 0
			cout << m << ":" << weight_[m] << "," << gauss_[m]->ComputeProb(input_vec) << "," << prob << endl;
#endif
		}
#if 0
		for (uint m = 0; m < gauss_.size(); ++ m) {
			prob += exp(tmp_[m] - max);
#if 0
			cout << prob << ", " << tmp_[m] << ", " << max << endl;
#endif
		}
#endif
#if 1
		return prob;
#else
		cout << log(prob) + max << endl;
		return (Rnum) (log(prob) + max);
#endif
	}
#else

	/*!
	 * @brief 引数の特徴ベクトルに対するGMMの尤度を計算する．
	 * @param Rnum *input_vec 尤度を計算する特徴量の系列へのポインタ
	 * @retval 入力の系列に対する対数尤度
	 */
	Rnum ComputeProb(Rnum *input_vec)
	{
		Rnum prob = -INFINITY;
		for (uint m = 0; m < gauss_.size(); ++ m) {
			prob = log_sum(prob, weight_[m] + gauss_[m]->ComputeProb(input_vec)); // log領域なので足し算
		}
		Rnum prob2 = 0.0, max = -INFINITY;
		tmp_.resize(gauss_.size());
		for (uint m = 0; m < gauss_.size(); ++ m) {
			tmp_[m] = weight_[m] + gauss_[m]->ComputeProb(input_vec);
			if (max < tmp_[m]) {
				max = tmp_[m];
			}
		}
		for (uint m = 0; m < gauss_.size(); ++ m) {
			prob2 += exp(tmp_[m] - max);
		}
		prob2 = log(prob2) + max;

		if (prob - prob2 != 0.0) {
			cout << prob << ", " << prob2 << ", " << prob - prob2 << endl;
		}
		return prob;
	}
#endif

	/*!
	 * @brief 引数の特徴ベクトルに対するGMMの尤度を計算する．
	 * @param Rnum *input_vec 尤度を計算する特徴量の系列へのポインタ
	 * @param int frame 尤度を計算するフレーム
	 * @retval 入力の系列に対する対数尤度
	 */
	Rnum ComputeProb(Rnum *input_vec, int frame)
	{
#if 1
		Rnum prob = -INFINITY;
#else
		Rnum prob = 0.0;
#endif
		for (int m = 0; m < (int) weight_.size(); ++ m) {
#if 1
			prob = log_sum(prob, weight_[m] + gauss_[m]->Score(input_vec, frame)); // log領域なので足し算
#else
			prob += weight_[m] * exp(gauss_[m]->Score(input_vec, frame) + kScale);
#endif
#if 0
			cout << m << ":" << weight_[m] << "," << gauss_[m]->Score(input_vec, frame) << "," << prob << endl;
#endif
		}
#if 1
		return prob;
#else
		return log(prob) - kScale;
#endif
	}

	/*!
	 * @brief 引数の特徴ベクトルに対するGMMの尤度を計算する．
	 * @param Rnum *input_vec 尤度を計算する特徴量の系列へのポインタ
	 * @retval 入力の系列に対する対数尤度
	 */
	Rnum ComputeProbScore(Rnum *input_vec)
	{
#if 1
		Rnum prob = -INFINITY;
#else
		Rnum prob = 0.0;
#endif
		for (int m = 0; m < (int) weight_.size(); ++ m) {
#if 1
			prob = log_sum(prob, weight_[m] + gauss_[m]->Score(input_vec)); // log領域なので足し算
#else
			prob += weight_[m] * exp(gauss_[m]->Score(input_vec) + kScale);
#endif
		}
#if 1
		return prob;
#else
		return log(prob) - kScale;
#endif
	}

	void set_mixture_(short mixture) { mixture_ = mixture; }
	void set_gauss_(const GaussContainer& gauss) { gauss_ = gauss; }
	void set_gauss_(Gauss *gauss, short m) { gauss_[m] = gauss; }
	void set_weight_(const WeightContainer& weight) { weight_ = weight; }
	void set_weight_(Rnum weight, short m) { weight_[m] = weight; }

	Gauss *get_gauss_(short m) { return gauss_[m]; }

	Rnum Score(Rnum *data, int frame) {
#ifdef SCORE_DEBUG
		cout << "GMM " << frame << " start" << endl;
#endif
		if (likelihood_[frame] > 0.0) {
			likelihood_[frame] = ComputeProb(data, frame);
		}
#ifdef SCORE_DEBUG
		cout << "GMM " << frame << " : " << likelihood_[frame] << endl;
#endif
		return likelihood_[frame];
	}

	Rnum Score(Rnum *data) {
		if (p_likelihood_ > 0.0) {
			p_likelihood_ = ComputeProbScore(data);
		}
		return p_likelihood_;
	}

	void Init(int sample_size) {
		likelihood_.resize(sample_size);
		for (int i = 0; i < sample_size; ++ i) {
			likelihood_[i] = 1.0;		// 未使用領域は正の値に初期化
		}
	}
	
	void Init() { p_likelihood_ = 1.0; }

	/*!
	 * @brief 使用されていないGauss分布の領域を開放して敷き詰める
	 * @retval void
	 */
	void RemoveEmptyGauss();

private:
	short mixture_;			///< 混合数（実際に格納されるガウス分布の数とは異なる場合がある）
	WeightContainer weight_;	///< 混合重み
	GaussContainer gauss_;		///< ガウス分布へのポインタを格納
	std::vector<Rnum> likelihood_;	///< 一度計算した値を保持しておくための領域
	std::vector<Rnum> tmp_;
	Rnum p_likelihood_;		///< 今のフレームで一度計算した値を保持しておくための領域（全フレーム覚えておくとメモリ喰いすぎる．．）
};

class StreamGmm {
public:
	typedef std::vector<Gmm *> GmmContainer;
	typedef std::vector<Rnum> WeightContainer;
	
	StreamGmm() : p_likelihood_(1.0) {}
	StreamGmm(short stream_num) : p_likelihood_(1.0) {
		gmm_.resize(stream_num);
	}
	virtual ~StreamGmm(){}

	Rnum ComputeProb(Rnum *data) {
		Rnum prob = 0.0;
		for (uint s = 0; s < gmm_.size(); ++ s) {
			prob += weight_->at(s) * gmm_[s]->ComputeProb(data);
#if 0
			cout << prob << endl;
#endif
		}
		return prob;
	}

	Rnum ComputeProb(Rnum *data, int frame) {
		Rnum prob = 0.0;
		for (uint s = 0; s < gmm_.size(); ++ s) {
			prob += w_vec_[s] * gmm_[s]->Score(data, frame);
#if 0
			cout << prob << endl;
#endif
		}
		return prob;
	}

	Rnum ComputeProbScore(Rnum *data) {
		Rnum prob = 0.0;
		for (uint s = 0; s < gmm_.size(); ++ s) {
			prob += w_vec_[s] * gmm_[s]->Score(data);
#if 0
			cout << prob << endl;
#endif
		}
		return prob;
	}
	
	Rnum Score(Rnum *data, int frame) {
#ifdef SCORE_DEBUG
		cout << "StreamGMM " << frame << " start" << endl;
#endif
		if (likelihood_[frame] > 0.0) {
#if 0
			likelihood_[frame] = ComputeProb(data, frame);
#else
			likelihood_[frame] = ComputeProb(data);
#endif
		}
#ifdef SCORE_DEBUG
		cout << "StreamGMM " << frame << " : " << likelihood_[frame] << endl;
#endif
		return likelihood_[frame];
	};

	Rnum Score(Rnum *data) {
		if (p_likelihood_ > 0.0) {
			p_likelihood_ = ComputeProbScore(data);
		}
		return p_likelihood_;
	};

	void Init(int sample_size);
	void Init() { p_likelihood_ = 1.0; }

	void set_gmm_(GmmContainer& gmm) { gmm_ = gmm; }
	void set_gmm_(Gmm *gmm, short s) { gmm_[s] = gmm; }
	void set_weight_(WeightContainer *weight) { weight_ = weight; w_vec_ = &(weight_->at(0)); }
	void set_weight_(Rnum weight, short s) { weight_->at(s) = weight; }
	Gmm *get_gmm_(short s) { return gmm_[s]; };

private:
	GmmContainer gmm_;
	WeightContainer *weight_;
	Rnum *w_vec_;
	std::vector<Rnum> likelihood_;
	Rnum p_likelihood_; 		///< 今のフレームで一度計算した値を保持しておくための領域（全フレーム覚えておくとメモリ喰いすぎる．．）
};

}
}

#endif // #ifndef _GMM_
