#ifndef _LANG_MODEL_
#define _LANG_MODEL_

/*!
 * @brief 言語モデル関係のプログラムのヘッダファイル達
 * 			@author Fujii Yasuhisa
 *			@date 29/1/2009
 */

#include <iostream>
#include <string>
#include <cstdlib>
#include <vector>
#include <set>

#include "my_typedefs.hpp"
#include "dictionary.hpp"

namespace slp {
namespace spojus {
using namespace std;

class LangModel {
public:
	static const Rnum kMinProb = -99.0;

	typedef set<Word> WordSet;

	static const short kMaxNgram = 10;
	struct ngram_chain {
		Word word;
		Word len;
	};

	struct ngram_prob {
		Word word;
		Rnum prob;
		Rnum backoff;
	};

	struct Ngram {
		union ngram_union {
			ngram_chain n_chain;
			ngram_prob n_prob;
		} gram;
		Ngram *ptr;
	};

public:
	enum Model {kSpojus, kArpa};
	LangModel() : verbose_(1), unk_penalty_(0.0) { dic_ = NULL; }
	LangModel(Dictionary *dic) : verbose_(1), unk_penalty_(0.0) { dic_ = dic; }

	virtual ~LangModel()
	{
		if (ngram_buff_ != NULL) {
			for (int n = 0; n < ngram_; ++ n) {
				for (int m = 0; m <= n; ++ m) {
					free(ngram_buff_[n][m]);
				}
				free(ngram_buff_[n]);
			}
			free(ngram_buff_);
		}
#ifdef DEBUG_DESTRUCTOR
		cout << "LangModel was freed successfully."<< endl;
#endif
	};

	void SetType(Model type) { type_ = type; }
	void SetDictionary(Dictionary *dic) { dic_ = dic; }
	short GetNgram () { return ngram_; }
	ulong GetNgramSize (short N) {
//		cout << N << ":" << ngram_ << endl;
		if (N >= ngram_) { return 0; }
		return ngram_size_[N - 1];
	}

	/*!
	 * @brief 与えられたファイルをオープンして言語モデルを読み込む
	 * @param const string& lmfile 言語モデルのファイル名
	 * @retval void
	 */
	void ReadModel(const string& lmfile);
	void ReadModel(const string& lmfile, Dictionary *dic) {
		dic_ = dic;
		ReadModel(lmfile);
	}

	virtual void Show();
	virtual Rnum Prob(Word *, short = 0);
	virtual Rnum BackoffProb(Word *, short);

	/*!
	 * @brief Perplexity を計算する
	 * @param WordContainer& words 単語系列を格納したコンテナ
	 * @param uint unk_num 未知語の数（デフォルト=1）
	 * @param int start 実際に確率を計算する単語の開始点（開始記号などに対処するため）
	 * @retval Rnum Perplexity
	 */
	virtual Rnum Perplexity(WordContainer& words, uint unk_num=1, int start=1);
	virtual Rnum ProbNoBackoff(Word *, short = 0);

        /*!
         * @brief 尤度 を計算する
         * @param WordContainer& words 単語系列を格納したコンテナ
         * @param int start 実際に確率を計算する単語の開始点（開始記号などに対処するため）
         * @retval Rnum Perplexity
         */
        virtual Rnum Likelihood(WordContainer& words, int start);

	/*!
	 * @brief コンテキストに続く単語を返す
	 * @param Word *words コンテキスト系列を格納した配列へのポインタ
	 * @param short N コンテキストの次数
	 * @param Ngram **n_ptr コンテキストに続く単語が開始する箇所へのポインタへのポインタ
	 * @param Word *len 単語が続く数
	 * @retval void
	 */
	void GetWordsByContext(Word *words, short N, Ngram **n_ptr, Word *len);

	/*!
	 * @brief 単語認識用に辞書ファイルから言語モデルを作成する
	 * @retval void
	 */
	void MakeDicLangModel();

protected:
	/*!
	 * @brief N-gramモデルを読み込む
	 * @param istream &in_fs 言語モデルファイルへの参照
	 * @param short ngram N-gramモデルの次数
	 * @retval void
	 */
	void ReadNgram(istream &in_fs, short ngram);

	/*!
	 * @biref N-gramファイルを解析する
	 * @param istream &in_fs N-gramファイルへの参照
	 * @retval void
	 */
	virtual void ParseFile(istream &in_fs) {};

	/*!
	 * @brief 辞書からNgramを作る
	 * @retval void
	 */
	void MakeDicNgram();

	/*!
	 * @brief 単語IDをクラスIDに変換する
	 * @param Word *words 単語ID列
	 * @param short N 単語ID列の長さ
	 * @retval void
	 */
	void Number2CID(Word *words, short N) {
		for (short i = 0; i < N; ++ i) {
			words[i] = dic_->Number2CID(words[i]);
		}
	}


public:
        void set_max_ngram_(short max_ngram) {
                max_ngram_ = max_ngram;
        }
        void set_ngram_(short ngram) {
                ngram_ = ngram;
        }

        short get_max_ngram_() { return max_ngram_; }
        short get_ngram_() { return ngram_; }

	void set_verbose_(char verbose) { verbose_ = verbose; }

	void set_unk_penalty_(Rnum unk_penalty) { unk_penalty_ = unk_penalty; }

public:
	short max_ngram_;	///< N-gramの最大次数
	short ngram_;		///< N-gramの実際に使用する最大次数
	Word words_[kMaxNgram];	///< 単語列記憶用領域

protected:
	static int CompareNgramStruct(const void *, const void *);
	static int CompareNgramStructNProb(const void *i, const void *j);

	char verbose_;
	Dictionary *dic_;
	vector<ulong> ngram_size_;
	Ngram ***ngram_buff_;
	Model type_;
	Rnum unk_penalty_;
};


class SpojusLangModel : public LangModel {
public:
	SpojusLangModel() { ngram_buff_ = NULL; type_ = kSpojus; }
	SpojusLangModel(Dictionary *dic) { dic_ = dic; type_ = kSpojus; }
	SpojusLangModel(const string &lmfile, char verbose_=1) { type_ = kSpojus; ReadModel(lmfile);};
	SpojusLangModel(const string &lmfile, Dictionary *dic, char verbose_=1) {
		dic_ = dic;
		type_ = kSpojus;
		ReadModel(lmfile);
	}

	virtual ~SpojusLangModel()
	{
#ifdef DEBUG_DESTRUCTOR
		cout << "SpojusLangModel was freed successfully."<< endl;
#endif
	};

private:
	/*!
	 * @biref Spojus形式のN-gramファイルを解析する
	 * @param istream &in_fs N-gramファイルへの参照
	 * @retval void
	 */
	void ParseFile(istream &in_fs);
};


class ArpaLangModel : public LangModel {
public:
	ArpaLangModel() { ngram_buff_ = NULL; type_ = kArpa; }
	ArpaLangModel(Dictionary *dic) { dic_ = dic; type_ = kArpa; }
	ArpaLangModel(const string &lmfile, char verbose_=1) { ReadModel(lmfile); type_ = kArpa; }
	ArpaLangModel(const string &lmfile, Dictionary *dic, char verbose=1) {
		verbose_ = verbose;
		dic_ = dic;
		type_ = kArpa;
		ReadModel(lmfile);
	}

	virtual ~ArpaLangModel()
	{
#ifdef DEBUG_DESTRUCTOR
		cout << "ArpaLangModel was freed successfully."<< endl;
#endif
	};

	virtual Rnum Prob(Word *words, short N)
	{
		if (N == 0) {
			N = ngram_;
		}
		if (dic_->class_num_ != 0) {
			for (short i = 0; i < N; ++ i) {
				words_[i] = words[i];
			}
			Number2CID(words_, N);
			return LangModel::Prob(words_, N) + dic_->GetDicWordFromWordId(words[N - 1])->class_score_;
		} else {
#if 0
			if (words[N - 1] == dic_->unk_id_) {
				cout << unk_penalty_ << endl;
			}
#endif
			return LangModel::Prob(words, N) + (words[N - 1] == dic_->unk_id_ ? -unk_penalty_ : 0.0);
		}
	}

private:
	/*!
	 * @biref Arpa形式のN-gramファイルを解析する
	 * @param istream &in_fs N-gramファイルへの参照
	 * @retval void
	 */
	void ParseFile(istream &in_fs);
};

}
}

#endif //ifndef _LANG_MODEL_
