#ifndef _LATTICE_
#define _LATTICE_

/**
 * @brief HMMトレリスからLatticeを生成するためのヘッダファイル
 *                         @author Fujii Yasuhisa
 *                         @date 28/1/2009
 */

#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <deque>

#include "spojus.hpp"
#include "asr_dictionary.hpp"

namespace slp {
namespace spojus {

class Lattice {
public:
	typedef BeamWord::BeamWordContainer BeamWordContainer;
	typedef BeamWord::BeamWordContainerContainer BeamWordContainerContainer;
	typedef BeamWord::BeamWordContainerContainerContainer BeamWordContainerContainerContainer;

	class Node;
	class Arc {
	public:
		Arc() :
			score_(-INFINITY),
			ac_score_(-INFINITY),
			lm_score_(-INFINITY),
			posterior_(-INFINITY),
			ac_posterior_(-INFINITY),
			snode_(NULL),
			enode_(NULL),
			stime_(-1),
			etime_(-1),
			asr_dic_word_(NULL),
			onebest_(false) {}

		~Arc() {
			if (asr_dic_word_ != NULL) {
				delete asr_dic_word_;
			}
		}

		Rnum score_;
		Rnum ac_score_;
		Rnum lm_score_;
		Rnum posterior_;
		Rnum ac_posterior_;
		Node *snode_;
		Node *enode_;
		int stime_;
		int etime_;
		AsrDicWord *asr_dic_word_;
		bool onebest_;
	};
	typedef deque<Arc *> ArcContainer;

	struct ArcCompare {
		bool operator() (Arc *arc1, Arc *arc2) const { 
			return arc1->posterior_ > arc2->posterior_;
		}
	};

	class Node {
	public:
		enum NodeState {kInitial, kFinal, kNormal};
		Node() : 
			forward_(-INFINITY), 
			backward_(-INFINITY),
			ac_forward_(-INFINITY),
			ac_backward_(-INFINITY),
			id_(-1)
#if 0
			container_(NULL)
			beam_word_(NULL)
#endif
		{
			type_ = kNormal;
		}

		void Init() {
			forward_ = -INFINITY;
			backward_ = -INFINITY;
			ac_forward_ = -INFINITY;
			ac_backward_ = -INFINITY;
		}

		ArcContainer starts_;
		ArcContainer ends_;
		Rnum forward_;
		Rnum backward_;
		Rnum ac_forward_;
		Rnum ac_backward_;
		NodeState type_;
		int time_;
		int id_;
#if 0
		BeamWordContainer *container_;
		BeamWord *beam_word_;	///< このノードに対応する単語終端へのポインタ
#endif
	};
	typedef deque<Node *> NodeContainer;
	
	Lattice(Spojus *spojus) {
		spojus_ = spojus;
		Init();
	};

	/*!
	 * @brief BeamWordからラティスを作るコンストラクタ
	 * @param BeamWordContainerContainerContainer& beam_word_ccc
	 * @param size_t beam_word_size 入力音声のフレーム数
	 * @param Spojus *spojus
	 * @param bool conf_net Confusion Network作成用のスコアにするかどうか（デフォルト=false）
	 * @retval void
	 */
	Lattice(BeamWordContainerContainerContainer& beam_word_ccc, size_t beam_word_size, Spojus *spojus, bool conf_net=false);

	virtual ~Lattice()
	{
		if (lfile_ != NULL) {
			lfile_->close();
			delete lfile_;
		}
		ContainerFree<NodeContainer>(nodes_);
		ContainerFree<ArcContainer>(arcs_);
	};

	void ReadFile(string& lattice_file);

	bool ReadLattice();

	void OpenFile(string& lattice_file) {
		lfile_ = new ifstream;
		lfile_->open(lattice_file.c_str());
	}
	void CloseFile() { lfile_->close(); }
	

	virtual void Show(); // Latticeの表示

	// アクセッサ
	int GetNodeNum(){ return nodes_.size(); }
	int GetArcNum(){ return arcs_.size(); }
	float GetVersion(){ return version_; }

	void set_mode_(int mode) { mode_ = mode; }

	/*!
	 * @brief Latticeの各アークについて，閾値を下回るアークを枝刈りする
	 * @param Rnum pthresh 枝刈りする閾値
	 * @retval void
	 */
	void Pruning(Rnum pthresh);

	/*!
	 * @brief アークを割合で枝刈りする
	 * @param Rnum rate 枝刈りする割合
	 * @retval void
	 */
	void PruningByRate(Rnum rate);

	void RegisterArc();

	Spojus *get_spojus_() { return spojus_; }

#if 0
	/*!
	 * @brief backwardをheuristcsとする探索のために，backwardスコアを最尤スコアに基づくものに変更する
	 * @retval void
	 */
	void AssignHeuristics();
#endif

	void set_conf_net_(bool conf_net) { conf_net_ = conf_net; }

	/*!
	 * @brief 単語トレリスからラティスを構築する
	 * @param BeamWordContainerContainerContainer& beam_word_ccc 単語トレリス
	 * @param size_t beam_word_size 入力音声のフレーム数
	 * @retval void
	 */
	void Build(BeamWordContainerContainerContainer& beam_word_ccc, size_t beam_word_size);

	/*!
	 * @brief Confusion Networkのためにスコア変換する
	 * @retval void
	 */
	void ReComputeScoreForConfNet();

protected:
	void ParseLine(string& );
	void RegisterArcs();
	void Init();

	/*!
	 * @brief 与えられた数だけバックトレースして単語を取得する
	 * @param BeamWord *beam_word バックトレースを開始する単語
	 * @param int ngram 単語を取得する数
	 * @retval void
	 *
	 * 取得した単語は lm_words_ に格納されます
	 *
	 */
	void GetLmWords(BeamWord *beam_word, int ngram);
	
private:
	/*!
	 * @brief 単語トレリス中の各単語について最終状態に至るかどうかをチェックする
	 * @param BeamWordContainerContainerContainer& beam_word_ccc
	 * @param size_t beam_word_size 入力音声のフレーム数
	 * @retval void
	 */
	void CheckReachable(BeamWordContainerContainerContainer& beam_word_ccc, size_t beam_word_size);
	/*!
	 * @brief 単語トレリスから可能な全てのパスを抽出し，ラティスを構成する．
	 * @param BeamWordContainerContainerContainer& beam_word_ccc
	 * @param size_t beam_word_size 入力音声のフレーム数
	 * @retval void
	 */
	void LatticeFromTrellis(BeamWordContainerContainerContainer& beam_word_ccc, size_t beam_word_size);
	void ComputeForward(int mode);
	void ComputeBackward(int mode);
	void ComputePosterior();

	/*!
	 * @brief BeamWordとCellからアークを作る
	 * @param BeamWord *beam_word アークの終端および単語の情報を保持
	 * @param Cell& diff_cell アークを作るコンテキストの情報を格納したCell
	 * @retval Arc * 作成したアークへのポインタ
	 */
	Arc *MakeArc(BeamWord *beam_word, Cell& diff_cell);

	/*!
	 * @brief 与えられた単語に対する最大の言語尤度を計算する
	 * @param Word word 言語尤度を計算する単語
	 * @param BeamWord *beam_word バックトレースを開始する単語
	 * @param Rnum ac_score 音響尤度
	 * @param int ngram バックトレースする単語数
	 * @retval void
	 */
	Rnum MaxLmProb(Word word, BeamWord *beam_word, Rnum ac_score, int ngram);
	
	/*!
	 * @brief 与えられた単語と単語履歴に対する言語尤度を計算する
	 * @param BeamWord *beam_word 言語尤度を計算する単語
	 * @param BeamWord *prev_beam_word バックトレースを開始する単語
	 * @param int ngram Nグラムの次数
	 * @retval void
	 */
	Rnum LmProb(BeamWord *beam_word, BeamWord *prev_beam_word, int ngram);


public:
	Word lm_words_[10];
	NodeContainer nodes_;
	ArcContainer arcs_;

protected:
	float version_;
	int utterance_;

private:
	Spojus *spojus_;
	ifstream *lfile_;
	enum SlfState {kHeader, kNode, kArc};
	SlfState slf_state_;
	bool conf_net_;					///< Confusion Network作成用のスコアにするかどうかのフラグ
	int mode_;					///< ForwardとBackward計算の時にViterbiにするか和にするか（0:viterbi, 1:和）
};

}
}

#endif // #ifndef _LATTICE_
