#ifndef _MY_MATH_
#define _MY_MATH_

#include <cmath>

namespace slp {
namespace spojus {

template<class T> T log_sum (T a, T b);
template<class T> T log_sub (T a, T b);

template<class T> T calc_determinant(T *matrix, ushort dimension);
template<class T> void calc_inverse(T *matrix, T *matrix_inv, ushort dimension);


/*!
 * @brief           
 * @param double *frame_data               
 * @param int size       
 * @retval double log energy
 */
template<class T>
double ComputeLogEnergy(T *frame_data, int size)
{
	double energy(0.0);
	for (int i = 0; i < size; ++ i) {
		energy += ((double) frame_data[i]) * ((double) frame_data[i]);
	}
	return (double) log(energy);
}


/*!
 * @brief Subtruct mean from data
 * @param T *data data sequence
 * @param int size data size
 * @retval void
 */
template<class T>
void ZeroMean(T *data, int size)
{
	int i;
	T ave(0.0);
	for (i = 0; i < size; ++ i) {
		ave += data[i];
	}
	ave /= (T) size;
	for (i = 0; i < size; ++ i) {
		data[i] -= ave;
	}
}


/*!
 * @brief ハミング窓をかける
 * @param T *frame_data ハミング窓をかけるデータ
 * @param int size
 * @retval void
 */
template<class T>
void HammingWindow(T *frame_data, int size)
{
	for (int i = 0; i < size; ++ i) {
		frame_data[i] *= 0.54 - 0.46 * cos(2.0 * M_PI * (double) i / (double) (size - 1));
	}
}

template<class T>
void make_sintbl(int n, T sintbl[])
{
	int i, n2, n4, n8;
	T c, s, dc, ds, t;

	n2 = n / 2;  n4 = n / 4;  n8 = n / 8;
	t = sin(M_PI / n);
	dc = 2 * t * t;  ds = sqrt(dc * (2 - dc));
	t = 2 * dc;  c = sintbl[n4] = 1;  s = sintbl[0] = 0;
	for (i = 1; i < n8; i++) {
		c -= dc;  dc += t * c;
		s += ds;  ds -= t * s;
		sintbl[i] = s;  sintbl[n4 - i] = c;
	}
	if (n8 != 0) sintbl[n8] = sqrt(0.5);
	for (i = 0; i < n4; i++)
		sintbl[n2 - i] = sintbl[i];
	for (i = 0; i < n2 + n4; i++)
		sintbl[i + n2] = - sintbl[i];
}

void make_bitrev(int n, int bitrev[]);

/*!
 * @brief FFT from "C言語によるアルゴリズム辞典"
 */
template<class T> 
int fft(int n, T x[], T y[])
{
	static int    last_n = 0;     
	static int    *bitrev = NULL; 
	static T *sintbl = NULL;
	int i, j, k, ik, h, d, k2, n4, inverse;
	T s, c, dx, dy;

	if (n < 0) {
		n = -n;  inverse = 1;
	} else     inverse = 0;
	n4 = n / 4;
	if (n != last_n || n == 0) {
		last_n = n;
		if (sintbl != NULL) delete sintbl;
		if (bitrev != NULL) delete bitrev;
		if (n == 0) return 0;
		sintbl = new T[n + n4];
		bitrev = new int[n];
		if (sintbl == NULL || bitrev == NULL) {
			std::cout << "Couldn't allocate memory" << std::endl;
			return 1;
		}
		make_sintbl(n, sintbl);
		make_bitrev(n, bitrev);
	}

	for (i = 0; i < n; i++) {
		j = bitrev[i];
		if (i < j) {
			std::swap(x[i], x[j]);
			std::swap(y[i], y[j]);
		}
	}

	for (k = 1; k < n; k = k2) {
		h = 0;  k2 = k + k;  d = n / k2;
		for (j = 0; j < k; j++) {
			c = sintbl[h + n4];
			if (inverse) s = - sintbl[h];
			else         s =   sintbl[h];
			for (i = j; i < n; i += k2) {
				ik = i + k;
				dx = s * y[ik] + c * x[ik];
				dy = c * y[ik] - s * x[ik];
				x[ik] = x[i] - dx;  x[i] += dx;
				y[ik] = y[i] - dy;  y[i] += dy;
			}
			h += d;
		}
	}
	if (inverse)
		for (i = 0; i < n; i++) { x[i] /= n;  y[i] /= n; }
	return 0; 
}


/*!
 * @brief 対数で表現された確率の和を対数領域で計算する
 * @param a 対数確率値1
 * @param b 対数確率値2
 * @retval T 計算された対数確率
 */
template<class T> T log_sum (T a, T b)
{
	if (a != -INFINITY && b != -INFINITY) {
		T e = exp(a - b);
		if (e == INFINITY) { return a > b ? a : b; }
		else { return b + log(1 + e); }
	}
	else if (a == -INFINITY) { return b; }
	else if (b == -INFINITY) { return a; }
	else { return -INFINITY; }
	
#if 0
	else if (a == b) {
		return a + log(2);
	}
	T sum = b + log(1 + exp(a - b));
	if (sum == INFINITY) { return a > b ? a : b; }
	return sum;
#endif
}


/*!
 * @brief 対数で表現された確率の差を対数領域で計算する
 * @param a 対数確率値1
 * @param b 対数確率値2
 * @retval T 計算された対数確率
 */
template<class T> T log_sub (T a, T b)
{

	if (a == -INFINITY && b == -INFINITY) {
		return -INFINITY;
	}

	if (a == -INFINITY) {
		return -INFINITY;
	}

	if (b == -INFINITY) {
		return a;
	}

	if (a <= b) {
		return -INFINITY;
	}
		
	return a + log(1 - exp(b - a));
}


/*!
 * @brief 行列の行列式を計算する
 * @param T *matrix 
 * @param ushort dimension
 * @retval T 
 */
template<class T> T calc_determinant(T *matrix, ushort dimension)
{
	int i, j, k;
	int i_index;
	int j_index;
	T buff;

	for (i = 0; i < dimension; i ++) {
		i_index = i * dimension;
		for (j = 0; j < dimension; j ++) {
			j_index = j * dimension;
			if (i < j) {
				buff = matrix[j_index + i] / matrix[i_index + i];

				for (k = 0; k < dimension; k ++) {
					matrix[j_index + k] -= matrix[i_index + k] * buff;
				}
			}
		}
	}

#if 0
	T determinant = 1.0;
	for (i = 0; i < dimension; i ++) {
		determinant *= matrix[i * dimension + i];
	}
#else
	T determinant = 0.0;
	for (i = 0; i < dimension; i ++) {
		determinant += log(matrix[i * dimension + i]);
	}
#endif
	return determinant;
}


/*!
 * @brief 行列の逆行列を求める
 *
 * 掃き出し法
 * 
 * @param[in] T *matrix 
 * @param[out] T *matrix_inv
 * @param ushort dimension[in]
 * @retval void
 */
template<class T> void calc_inverse(T *matrix, T *matrix_inv, ushort dimension)
{
	int i, j, k;
	int i_index, j_index;
	T buff;

	for (i = 0; i < dimension; i ++) {
		i_index = i * dimension;
		for (j = 0; j < dimension; j ++) {
			matrix_inv[i_index + j] = (i == j ? 1.0 : 0.0);
		}
	}

	for (i = 0; i < dimension; i ++) {
		i_index = i * dimension;
		buff = matrix[i_index + i];
		for (j = 0; j < dimension; j ++) {
			matrix[i_index + j] /= buff;
			matrix_inv[i_index + j] /= buff;
		}

		for (j = 0; j < dimension; j ++) {
			j_index = j * dimension;

			if (i == j) {
				continue;
			}

			buff = matrix[j_index + i];
			for (k = 0; k < dimension; k ++) {
				matrix[j_index + k] -= matrix[i_index + k] * buff;
				matrix_inv[j_index + k] -= matrix_inv[i_index + k] * buff;
			}
		}
	}
	
}

}
}

#endif // #ifndef _MY_MATH_
