#ifndef _ONE_PASS_DP_
#define _ONE_PASS_DP_
/*!
 * @brief OnePassDP法によって1パス目の認識結果を求めるプログラムのヘッダ
 * @author Fujii Yasuhisa
 * @date 29/04/2009
 */

#include <queue>
#include <vector>
#include <functional>

#include "my_typedefs.hpp"
#include "asr_dictionary.hpp"

namespace slp {
namespace spojus {

class Spojus;		//前方参照
class InputData;

class OnePassDP {
#if 1
	static const int kPreReserve = 10000;
#else
	static const int kPreReserve = 0;
#endif
public:
	typedef AsrDictionary::Node AsrNode;
	typedef AsrDictionary::NodePointer AsrNodePointer;
	typedef AsrDictionary::NodePointer::Greater0 AsrNodePointerGreater0;
	typedef AsrDictionary::NodePointer::Greater1 AsrNodePointerGreater1;
#if 1
	typedef AsrDictionary::NodePointer::WordOrder AsrNodePointerWordOrder;
	typedef AsrDictionary::NodePointer::WordNumberEqual AsrNodePointerWordNumberEqual;
#endif
	typedef AsrDictionary::Node::TransitionContainer AsrTransitionContainer;
	typedef vector<AsrDictionary::NodePointer *> NodePtrContainer;
	typedef BeamWord::BeamWordContainer BeamWordContainer;
	typedef BeamWord::BeamWordContainerContainer BeamWordContainerContainer;
	typedef BeamWord::BeamWordContainerContainerContainer BeamWordContainerContainerContainer;
	typedef BeamWord::Compare BeamWordCompare;

public:
	OnePassDP(){};
	OnePassDP(Spojus *spojus){
		spojus_ = spojus;
		ContainerReserve();
	};
	
	virtual ~OnePassDP(){
		Clear();
		// tmp_node_ptr_container_は後に再利用出来るように，デストラクタで開放する
		ContainerFree<NodePtrContainer>(tmp_node_ptr_container_);
#ifdef DEBUG_DESTRUCTOR
		cout << "OnePassDP was freed successfully."<< endl;
#endif
	};

	void Clear() {
		ContainerCutFree(t_initial_node_ptr_container_);
		ContainerCutFree(t_node_ptr_container_);
		ContainerCutFree(t_null_node_ptr_container_);
		ContainerCutFree(l_initial_node_ptr_container_);
		ContainerCutFree(l_node_ptr_container_);
		ContainerCutFree(l_null_node_ptr_container_);
		ContainerCutFree(word_node_ptr_container_);
		ContainerCutFree(linked_node_ptr_container_);
		// tmp_node_ptr_container_はCutLinkのみ
		CutLink(tmp_node_ptr_container_, tmp_node_ptr_container_.begin());
		// Free beam_word_
		BeamWordContainerContainerContainer::iterator bwccc_iter;
		BeamWordContainerContainer::iterator bwcc_iter;
		for (bwccc_iter = beam_word_.begin(); bwccc_iter != beam_word_.end(); ++ bwccc_iter) {
			for (bwcc_iter = bwccc_iter->begin(); bwcc_iter != bwccc_iter->end(); ++ bwcc_iter) {
				ContainerFree<BeamWordContainer>(*bwcc_iter);
			}
		}
	}

	void set_input_data_(InputData *input_data) {
		input_data_ = input_data;
	}

	/*!
	 * @brief サーチを始める前の初期化を行う
	 * @retval void
	 */
	void Init();

	
	/*!
	 * @brief 指定したフレームからバックトレースして，1best解を表示する
	 * @retval void
	 */
	void BackTrace(size_t frame);

        /*!
         * @brief beam_word_をリサイズする
         * @param int frame リサイズする大きさ
         * @retval void
         */
        void BeamWordResize(int frame);

	/*!
	 * @brief サーチの終了処理を行う
	 * @retval void
	 */
	void Finalize();

	/*!
	 * @brief 認識結果をXMLにして送信する
	 * @retval void
	 */
	void SendXML(MySocket *socket);

	/*!
	 * @brief OnePassDP法に基づく探索を行う
	 * @param InputData * input_data 入力データへのポインタ
	 * @retval void
	 */
	void Search(InputData *);

	/*!
	 * @brief OnePassDP法に基づく探索において，各フレームでの処理を行う
	 * @param int frame 処理を行うフレーム
	 * @retval void
	 */
	void SearchAtFrame(int frame);

	/*!
	 * @brief 1フレーム分ビーム格納領域を拡張する
	 * @retval void
	 */
	void ExpandBeamWord();

	/*!
	 * @brief 開始記号をビームに登録
	 * @retval void
	 */
	void AddBeginningBeamWord();

	/*!
	 * @brief 1best解の各単語終端でビームに残っていた単語を出力する
	 * @param const BeamWordContainer& best_seq 1best解の単語列を格納したコンテナ
	 * @retval void
	 */
	void Output1bestTrellis(const BeamWordContainer& best_seq);

	/*!
	 * @brief 1パス目のトレリスの各単語終端でビームに残っていた単語を出力する
	 * @param WordContainer& words 1best単語列
	 * @param vector<int>& bound 1best単語列の単語境界
	 * @retval void
	 */
	void Output1bestTrellis(WordContainer& words, vector<int>& bound);

	/*!
	 * @brief 指定したフレームでの最尤仮説のスコアを出力する
	 * @param const std::vector<int>& frames スコアを出力するフレーム
	 * @retval void
	 */
	void PrintBestScore(const std::vector<int>& frames);

private:
	enum DicType {kTrie, kLinear, kOther};			///< 辞書の種類

	/*!
	 * @brief コンテナに使用する領域を予め確保する
	 * @retval void
	 */
	void ContainerReserve()
	{
#if 0
		t_node_ptr_container_.reserve(kPreReserve);
		t_null_node_ptr_container_.reserve(kPreReserve);
		l_initial_node_ptr_container_.reserve(kPreReserve);
		l_node_ptr_container_.reserve(kPreReserve);
		l_null_node_ptr_container_.reserve(kPreReserve);
		word_node_ptr_container_.reserve(kPreReserve);
		linked_node_ptr_container_.reserve(kPreReserve);
#endif
		tmp_node_ptr_container_.reserve(kPreReserve);
	}


	/*!
	 * @brief コンテナに格納されたノードポインタを参照するリンクを削除するとともに，ノードポインタの領域を開放する
	 * @param NodePtrContainer& node_ptr_container 開放するコンテナ
	 * @retval void
	 */
	void ContainerCutFree(NodePtrContainer& node_ptr_container);

	/*!
	 * @brief 与えられたコンテナのデータを，リンクを切ったうえでtmp_node_cotainer_ptr_に移動して空にする
	 * @param NodePtrContainer& node_ptr_container データを空にしたいコンテナ
	 * @retval void
	 */
	void ContainerCutClear(NodePtrContainer& node_ptr_container);

	/*!
	 * @brief 与えられたコンテナの要素をtmp_node_ptr_container_に移動する
	 * @param NodePtrContainer& node_ptr_container データを移動したいコンテナ
	 * @retval void
	 */
	void ContainerClear(NodePtrContainer& node_ptr_container);

	/*!
	 * @brief ビームに残る単語を格納する領域の初期化
	 * @retval void
	 */
	void InitBeamWord();

	/*!
	 * @brief ノード数を出力する
	 * @param const string& message 出力の先頭に付加する文
	 * @retval
	 */
	void NodeNum(const string& message);

	/*!
	 * @brief ノードポインタに重複が無いかどうかチェック
	 * @param AsrNodePointer *node_ptr 重複をチェックするノードポインタ
	 * @retval void
	 */
	void ValidateNodePtr(AsrNodePointer *node_ptr);

	/*!
	 * @brief 単語の始端への遷移を計算する
	 * @param NodePtrContainer& node_ptr_container ノードポインタのコンテナ
	 * @param DicType dic 辞書の種類
	 * @retval void
	 */
	void InitialTransition(NodePtrContainer& node_ptr_container, DicType dic);

	/*!
	 * @brief 線形辞書の単語の始端への遷移を計算する
	 * @retval void
	 */
	void MakeLinearTransition(int frame);

	/*!
	 * @brief 線形辞書の単語の始端への遷移を計算する
	 * @retval void
	 */
	void LinearInitialTransition();

	/*!
	 * @brief 音響マッチングを行う
	 * @param NodePtrContainer& node_ptr_container ノードポインタのコンテナ
	 * @param Rnum *data 音響尤度を計算する特徴量
	 * @param int frame 現在のフレーム
	 * @param DicType dic 辞書の種類
	 * @retval void
	 */
	void AcousticMatching(NodePtrContainer& node_ptr_container, Rnum *data, int frame, DicType dic);

	/*!
	 * @brief Null遷移を計算する
	 * @param NodePtrContainer& node_ptr_container ノードポインタのコンテナ
	 * @param DicType dic 辞書の種類
	 * @param int mode 0 : 1 音響マッチングの前か後か
	 * @retval void
	 */
	void NullTransition(NodePtrContainer& node_ptr_container, DicType dic, int mode);

	/*!
	 * @brief 枝刈りを行う
	 * @retval void
	 */	
	void Pruning();

	/*!
	 * @brief スコアと最大ノード数を基準にリスト内のノードを枝刈りする
	 * @param NodePtrContainer& node_ptr_container ノードを格納するリストへの参照
	 * @param uint max_num 最大ノード数
	 * @param Rnum max_score 最大スコア
	 * @param int idx スコアを考慮するインデックス
	 * @retval void
	 */
	void PruneList(NodePtrContainer& node_ptr_container, uint max_num, Rnum max_score, int idx);

	/*!
	 * @brief スコアを基準に2つのリスト内のノードを枝刈りする
	 * @param NodePtrContainer& node_ptr_container1 ノードを格納するリスト1への参照
	 * @param NodePtrContainer& node_ptr_container2 ノードを格納するリスト2への参照
	 * @param Rnum top_score 枝刈りの基準となる値
	 */
	void PruneList(NodePtrContainer& node_ptr_container1, NodePtrContainer& node_ptr_container2, Rnum top_score);

	/*!
	 * @brief Hmm::NodePointerとHMM::Nodeの間のリンクを切る
	 * @param NodePtrContainer& node_ptr_container リンクを切りたいNodePointerを格納したコンテナへの参照
	 * @param NodePtrContainer::iterator npc_iter リンクを切りたい開始NodePointerへのiterator
	 * @retval void
	 */
	void CutLink(NodePtrContainer& node_ptr_container, NodePtrContainer::iterator npc_iter);

	/*!
	 * @brief ビーム幅に残る単語を選択
	 * @retval void
	 */
	void Beam(int frame);

	/*!
	 * @brief 言語モデル確率の再計算を行う
	 * @retval void
	 */
	void ReComputeLangModelProb();

	/*!
	 * @brief 与えられた単語で終端する最大のN-gram確率を計算し，スコアに加算する
	 * @param AsrNodePointer *node_ptr 単語ノードポインタへのポインタ
	 * @retval void
	 */
	void AddLangModelProb(AsrNodePointer *node_ptr);

	/*!
	 * @brief 与えられた単語履歴でのN-gram確率を計算し，スコアに加算する
	 * @param AsrNodePointer *node_ptr 単語ノードポインタへのポインタ
	 * @retval void
	 */
	void AddLangModelProbLinear(AsrNodePointer *node_ptr);

	/*!
	 * @brief 与えられた単語で終端する最大のN-gram確率を計算する（N>2の時は近似的となる）
	 * @param AsrNodePointer *node_ptr N-gram確率を計算したい単語ノードポインタへのポインタ
	 * @retval void
	 */
	void MaxLangModelProb(AsrNodePointer *node_ptr);

	/*!
	 * @brief 与えられた数だけバックトレースして単語を取得する
	 * @param BeamWord *beam_word バックトレースを開始する単語
	 * @param int ngram 単語を取得する数
	 * @retval void
	 *
	 * 取得した単語は lm_words_ に格納されます
	 *
	 */
	void GetLmWords(BeamWord *beam_word, int ngram);

	/*!
	 * @brief リンクされた単語からの遷移を計算する
	 * @retval void
	 */
	void LinkedNodeTransition();

	/*!
	 * @brief ノードのスコア及びバックポインタを更新する
	 * @param AsrNodePointer *node_ptr 更新するノード
	 * @param AsrNodePointer *from_ptr 伝搬元ノードポインタへのポインタ
	 * @param Rnum score 新しいスコア
	 * @param int from スコアの更新元
	 * @param int to スコアの更新先
	 * @retval void
	 */
	void UpdateNode(AsrNodePointer *node_ptr, AsrNodePointer *from_ptr, Rnum score, int from, int to);

	/*!
	 * @brief Diffノードのスコア及びバックポインタを更新する
	 * @param AsrNodePointer *node_ptr 更新するノード
	 * @param AsrNodePointer *from_ptr 伝搬元ノードポインタへのポインタ
	 * @param Rnum score 新しいスコア
	 * @param int from スコアの更新元
	 * @param int to スコアの更新先
	 * @retval void
	 */
	void UpdateDiffNode(AsrNodePointer *node_ptr, AsrNodePointer *from_ptr, Rnum score, int from, int to);

	/*!
	 * @brief ビームに残った単語を保存する
	 * @param int frame 現在のフレーム
	 * @retval void
	 */
	void StoreBeam(int frame);

	/*!
	 * @brief ビームに残った単語をinitialコンテナにpushする
	 * @param int frame 現在のフレーム
	 * @retval void
	 */
	void PushBeamWord(int frame);


	/*!
	 * @brief NodePtrContainerのスコアをローテーションする
	 * @retval void
	 */
	void Rotate();

	/*!
	 * @brief 引数で与えられたNodePtrContainerのスコアをローテーションする
	 * @param NodePtrContainer& node_ptr_container NodePtrContainerへの参照
	 * @retval void
	 */
	void RotateContainer(NodePtrContainer& node_ptr_container);

	/*!
	 * @brief ノードをアクティブにする
	 * @param DicType dic 辞書の種類
	 * @retval void
	 */
	AsrNodePointer *ActivateNode(AsrNode *node, DicType dic);

	/*!
	 * @brief BeamWordを表示する
	 * @retval void
	 */
	void ShowBeamWord();

	/*!
	 * @brief frame目のBeamWordを表示する
	 * @param frame BeamWordを表示するフレーム数
	 * @retval void
	 */
	void ShowBeamWord(int frame);

	/*!
	 * @brief リストをソートしてスコアの最大値を返す
	 * @param NodePtrContainer& node_ptr_container ノードポインタを格納したリストへの参照
	 * @retval Rnum リストに格納されているノードポインタのスコアの最大値
	 */
	Rnum GetTopScore(NodePtrContainer& node_ptr_container);

	/*!
	 * @brief バックトレースして，1best解を表示する
	 * @retval void
	 */
	void BackTrace();

	/*!
	 * @brief 最終状態への遷移を計算する
	 * @retval void
	 */
	void EndProb();

	/*!
	 * @brief 現在のフレームでの最尤仮説のスコアを保存する
	 * @param const NodePtrContainer& node_ptr_container 各コンテキストでの最尤ノードを格納したコンテナ
	 * @param const int frame 現在のフレーム
	 * @retval void
	 */
	void SaveBestScore(const NodePtrContainer& node_ptr_container, const int frame);

public:
	Spojus *spojus_;						///< Spojusクラスへのポインタ
	InputData *input_data_;						///< 入力データ格納領域へのポインタ
	size_t beam_word_size_;						///< 入力データのフレーム数
	BeamWordContainerContainerContainer beam_word_;			///< ビームに残った単語を保存する変数
	Word lm_words_[10];						///< 言語モデル計算用単語列記憶領域

	StringContainer one_best_;					///< 認識結果格納用領域
	StringContainer& get_one_best_() {return one_best_;}		///< アクセッサ

private:
	NodePtrContainer t_initial_node_ptr_container_;			///< 単語の始端に接続されるノードへのポインタ
	NodePtrContainer l_initial_node_ptr_container_;			///< 線形辞書の単語の始端に接続されるノードへのポインタ
	NodePtrContainer t_node_ptr_container_;				///< アクティブノードへのポインタ（木辞書）
	NodePtrContainer t_null_node_ptr_container_;			///< Null遷移を持つアクティブノードへのポインタ（木辞書）
	NodePtrContainer l_node_ptr_container_;				///< アクティブノードへのポインタ（線形辞書）
	NodePtrContainer l_null_node_ptr_container_;			///< Null遷移を持つアクティブノードへのポインタ（線形辞書）
	NodePtrContainer word_node_ptr_container_;			///< 単語ノードのアクティブノードへのポインタ
	NodePtrContainer linked_node_ptr_container_;			///< リンクされた単語ノードへのポインタ
	NodePtrContainer tmp_node_ptr_container_;			///< ノード再利用のための一時保存場所

	std::vector<int> best_frame_scores_;

	struct TimeCount {
		TimeCount() { Init(); }

		void Init() { 
			init_ = 0;
			linear_initial_transition_ = 0;
			initial_transition_ = 0;
			null_transition_ = 0;
			acoustic_matching_ = 0;
			gmm_computing_ = 0;
			beam_ = 0;
			recompute_langmodel_prob_ = 0;
			lm_computing_ = 0;
			pruning_ = 0;

			rotate_ = 0;
		}

		void Show() {
                        cout.setf(ios_base::fixed);
                        cout.precision(3);
			cout << "Init                       : "
                             << (double) init_ / CLOCKS_PER_SEC << endl;
			cout << "Linear Initial transition  : "
                             << (double) linear_initial_transition_ / CLOCKS_PER_SEC << endl;
			cout << "Initial transition         : "
                             << (double) initial_transition_ / CLOCKS_PER_SEC << endl;
			cout << "Null transition            : "
                             << (double) null_transition_ / CLOCKS_PER_SEC << endl;
			cout << "Acoustic matching          : "
                             << (double) acoustic_matching_ / CLOCKS_PER_SEC << endl;
#ifdef CALC_TIME_IN
			cout  << "\tgmm="
                              << (double) gmm_computing_ / CLOCKS_PER_SEC
                              << ", other=" 
                              << (double) (acoustic_matching_ - gmm_computing_)
                                / CLOCKS_PER_SEC << endl;
#endif
			cout << "Beam                       : "
                             << (double) beam_ / CLOCKS_PER_SEC << endl;
#ifdef CALC_TIME_IN 
			cout << "\tlm="
                             << (double) lm_computing_ / CLOCKS_PER_SEC
                             << ", other="
			     << (double) beam_ - lm_computing_
                                / CLOCKS_PER_SEC << endl;
#endif
			cout << "Pruning                    : "
                             << (double) pruning_ / CLOCKS_PER_SEC
                             << endl;
			cout << "Rotate                     : "
                             << (double) rotate_ / CLOCKS_PER_SEC << endl;
		}

		clock_t init_;
		clock_t linear_initial_transition_;
		clock_t initial_transition_;
		clock_t acoustic_matching_;
		clock_t null_transition_;
		clock_t gmm_computing_;
		clock_t beam_;
		clock_t recompute_langmodel_prob_;
		clock_t lm_computing_;
		clock_t pruning_;
		clock_t rotate_;
	};
	TimeCount time_count_;
};

}
}

#endif // #ifndef _ONE_PASS_DP_
