#ifndef _OPTIONS_
#define _OPTIONS_
/*!
 * @brief オプションを処理するプログラム
 * 受け取ったオプションは全てOptionsクラスに格納する
 * @author Fujii Yasuhisa
 * @date 04/28/2009
 */

#include <iostream>
#include <list>
#include <vector>
#include <string>
#include <cstdlib>

#include <netinet/in.h>

#include "my_typedefs.hpp"

namespace slp {
namespace spojus {
using namespace std;

class AcousticModel;
class LangModel;
class Dictionary;

template <class T>
struct ModelPack {
	T model_;
	string file_;
};

class Options {
public:
#ifdef HCNF
	static const float kHCNFAlpha = 0.1;
	static const float kHCNFBeta = 0.0;
	static const float kHCNFSigAmp = 6.0;
	static const float kHCNFSigBias = 3.0;
	static const float kHCNFKappa = 0.1;
#endif
	struct LmWeight {
		int index_;
		Rnum weight_;
	};
	typedef vector<LmWeight> LmWeightContainer;

	Options() : unk_switch_(true) { Init(); };
	Options(int args_num, char *args[]) : unk_switch_(true) {
		Init();
		if (false == ParseCommand(args_num, args)) {
			cerr << "Failed to parse command line argments.." << endl;
			exit(1);
		}
	}

	virtual ~Options(){};

	/*!
	 * @brief オプション変数に初期値を与える
	 * @retval void
	 */
	void Init()
	{
		left_initial_context_	= "SIL";		///< 左の開始コンテキスト
		right_ending_context_	= "SIL";		///< 右の終端コンテキスト
		max_node_num_		= 3000;			///< 通常ノードおよびNullノードそれぞれで最大保持するノード数
		linear_max_node_num_	= 500;			///< 線形辞書の通常ノードおよびNullノードそれぞれで最大保持するノード数
		beam_width_		= 30;			///< ビーム幅
		beam_thr_		= 400;			///< 枝刈りのしきい値
		show_beam_		= 0;			///< ビームに残った単語を表示するかどうか
		roll_back_		= false;		///< 1パス目の最終フレームで仮設が見つからなかった時に遡って仮説を見つけるかどうか
		word_penalty_ 		= 0.0;			///< 単語の挿入ペナルティ
		syll_penalty_ 		= 0.0;			///< 音節の挿入ペナルティ
		state_penalty_ 		= 0.0;			///< 状態の遷移ペナルティ
		lm_weight_		= 1.0;			///< 言語重み	
		first_ngram_		= 0;			///< 第1パスで使用する言語モデルの次数 
		second_ngram_		= 0;			///< 第2パスで使用する言語モデルの次数
		factoring_		= 1;			///< Unigram Factoringを行うかどうか
		linear_num_		= 0;			///< 線形辞書の本数
		nbest_num_		= 0;			///< 2パス目で求める仮説数
		second_beam_width_      = 1000;			///< 2パス目でのビーム幅（各ステップで残す仮説数）
		swap_			= false;		///< 入力のswapを行うかどうか
		send_port_		= 30000;		///< 送信ポート
		receive_port_		= 30001;		///< 受信ポート
		client_port_		= 15920;		///< フロントエンド用ポート
		frontend_port_		= 15921;		///< フロントエンド用ポート
		send_			= false;		///< 認識結果をネットワーク越しに送信するかどうか
		send_host_		= "localhost";		///< 送信ホスト
		frontend_host_		= "localhost";		///< フロントエンドホスト
		feature_type_		= kAsIs;		///< 特徴量の並び
		conf_net_		= false;		///< コンフュージョンネットワークの構築を行うかどうか
		conf_net_lattice_pruning_ratio_	= 1.0;		///< コンフュージョンネットワークを作成する時のラティス枝刈りしきい値
		conf_net_size_pruning_	= true;			///< コンフュージョンネットワークの枝刈りをサンプル数に対する比で行うかどうか;
		output_1best_trellis_	= false;		///< 1best解の各単語終端でビームに残っていた単語を出力するかどうか
		editing_beam_width_	= 1000;			///< 整形処理でのビーム幅
		editing_		= false;		///< 整形処理を行うかどうか
		editing_translation_lm_index_	= -1;		///< 変換モデルのインデックス
		editing_posterior_weight_ = 3.0;		///< 整形処理におけるpositeriorの重み
		editing_del_prior_	= log(0.2);		///< 整形処理における脱落の事前確率
		editing_lm_weight_	= 1.0;			///< 整形処理における言語重み
		editing_insert_		= false;		///< 挿入を考慮するかどうか
		gmm_frame_number_	= 0;			///< 音響モデル選択に使用するフレーム数（0=全部）
		parm_kind_string_	= "MFCC_E_D_A_Z";	///< 音響モデルの特徴量
		cepstrum_dim_		= 12;			///< 生の特徴量の次元数
		dimension_		= 39;			///< 特徴量の次元数
		recog_mode_		= kNgram;		///< 認識モード
		recog_unit_		= kSyll;		///< 認識単位
		inter_word_right_context_	= false;	///< 単語間右コンテキスト処理を行うかどうか
		sepnum_			= 0;			///< 第1パスで木辞書から独立させる高頻度語の数（単語数）
		best_pronounciation_	= false;		///< 複数の発音が存在する単語について，最尤の単語だけ残す
		verbose_		= 1;			///< ログ出力レベル
		demo_			= false;		///< デモ出力モードか否か

		unk_penalty_		= 0.0;			///< 未知語ペナルティ
#ifdef HCNF
		hcnf_alpha_ = kHCNFAlpha;
		hcnf_beta_ = kHCNFBeta;
		hcnf_sigamp_ = kHCNFSigAmp;
		hcnf_sigbias_ = kHCNFSigBias;
		hcnf_kappa_ = kHCNFKappa;
#endif
	}

	/*!
	 * @brief オプションを表示する
	 * @retval void
	 */
	void Show();

	/*!
	 * @brief *argv[]を受け取ってオプションを解析する
	 * @param int args_num 引数の数
	 * @param char *args[] コマンドライン引数の*argv[]を受け取る変数
	 * @retval bool オプション解析の成否
	 */
	bool ParseCommand(int args_num, char *args[]);

	void set_unk_switch_(bool unk_switch) { unk_switch_ = unk_switch; }

private:
	/*!
	 * @brief 存在しないオプションに対する警告を出力する
	 * @param const string& opt;
	 * @retval void
	 */
	void UnknownOption(const string& opt);

	bool ParseConfig(const string &config_file);
	bool IsSkipCharactor(char c) {
		if (c == ' ' || c == '\t' || c == '=' || c == '"') return true;
		else return false;
	}

public:
	enum InputType {kList, kStdIn, kFile, kPcm, kMic, kServer, kKaiServer};
	enum RecogMode {kNgram, kCFG, kWordRecog};
	enum RecogUnit {kSyll, kPhone};

	InputType input_type_;
	list<string> input_file_list_;				// 入力ファイルのリストのリスト
	list<ModelPack<AcousticModel *> > am_file_;		// 音響モデルファイルのリスト
	list<ModelPack<LangModel *> > lm_file_;			// 言語モデルファイルのリスト
	list<ModelPack<Dictionary *> > dic_file_;		// 辞書ファイルのリスト
	list<string> gmm_file_;					// GMM定義ファイルのリスト
	string hlist_file_;					// HMMのリストを定義したファイル
	string class_file_;

	string left_initial_context_;				///< 左の開始コンテキスト
	string right_ending_context_;				///< 右の終端コンテキスト

	uint max_node_num_;					///< 通常ノードおよびNullノードそれぞれで最大保持するノード数
	uint linear_max_node_num_;				///< 線形辞書の通常ノードおよびNullノードそれぞれで最大保持するノード数
	ushort beam_width_;					///< ビーム幅
	Rnum beam_thr_;						///< 枝刈りのしきい値
	char show_beam_;					///< ビームに残った単語を表示するかどうか
	bool roll_back_;					///< 1パス目の最終フレームで仮設が見つからなかった時に遡って仮説を見つけるかどうか

	Rnum word_penalty_;					///< 単語の挿入ペナルティ
	Rnum syll_penalty_;					///< 音節の挿入ペナルティ
	Rnum state_penalty_;					///< 状態の遷移ペナルティ

	Rnum lm_weight_;					///< 言語重み
	list<LmWeightContainer> interpolation_;			///< 言語モデルインデックスと内挿重みのリスト

	short first_ngram_;					///< 第1パスで使用する言語モデルの次数 
	short second_ngram_;					///< 第2パスで使用する言語モデルの次数

	char factoring_;					///< Unigram Factoringを行うかどうか

	Word linear_num_;					///< 線形辞書の本数
	ushort nbest_num_;					///< 2パス目で求める仮説数
	ushort second_beam_width_;				///< 2パス目でのビーム幅（各ステップで残す仮説数）

	bool swap_;						///< 入力のswapを行うかどうか

	bool send_;						///< 認識結果をネットワーク越しに送信するかどうか
	in_port_t send_port_;					///< 送信ポート
	in_port_t receive_port_;				///< 受信ポート
	in_port_t client_port_;					///< クライアント用ポート
	in_port_t frontend_port_;				///< フロントエンド用ポート
	string send_host_;					///< 送信ホスト
	string frontend_host_;					///< フロントエンドホスト

	enum FeatureType {kBlock, kRmRaw, kAsIs};		///< 特徴量の並びのタイプ
	FeatureType feature_type_;				///< 特徴量の並び

	bool conf_net_;						///< コンフュージョンネットワークの構築を行うかどうか
	Rnum conf_net_lattice_pruning_ratio_;			///< コンフュージョンネットワークを作成する時のラティス枝刈りしきい値
	bool conf_net_size_pruning_;				///< コンフュージョンネットワークの枝刈りをサンプル数に対する比で行うかどうか

	bool output_1best_trellis_;				///< 1best解の各単語終端でビームに残っていた単語を出力するかどうか

	ushort editing_beam_width_;				///< 整形処理でのビーム幅
	short editing_translation_lm_index_;			///< 変換モデルのインデックス
	bool editing_;						///< 整形処理を行うかどうか
	Rnum editing_posterior_weight_;				///< 整形処理におけるpositeriorの重み
	Rnum editing_del_prior_;				///< 整形処理における脱落の事前確率
	Rnum editing_lm_weight_;				///< 整形処理における言語重み
	bool editing_insert_;					///< 挿入を考慮するかどうか

	string alignment_list_;					///< アライメントを行う単語列のリストを格納したファイル名

	list<string> option_box_;				///< 他のプログラムでoptionを処理できるようにするための避難領域

	string ccs_file_;					///< コンテキストキューを格納したファイル

        int gmm_frame_number_;					///< 音響モデル選択に使用するフレーム数
	string parm_kind_string_;				///< 音響モデルの特徴量
	short cepstrum_dim_;					///< 生の特徴量の次元数
	int dimension_;						///< 特徴量の次元数

	RecogMode recog_mode_;					///< 認識モード
	RecogUnit recog_unit_;					///< 認識単位
	bool inter_word_right_context_;				///< 単語間右コンテキスト処理を行うかどうか
	
	Word sepnum_;						///< 第1パスで木辞書から独立させる高頻度語の数（単語数）
	bool best_pronounciation_;				///< 複数の発音が存在する単語について，最尤の単語だけ残す

	char verbose_;						///< ログ出力レベル
	bool demo_;						///< デモ出力モードか否か;

	Rnum unk_penalty_;					///< 未知語ペナルティ

	std::vector<int> best_score_print_frames_;		///< 最尤スコアを出力するフレーム

#ifdef HCNF
	string hcnf_model_file_;
	float hcnf_alpha_;
	float hcnf_beta_;
	float hcnf_sigamp_;
	float hcnf_sigbias_;
	float hcnf_kappa_;
#endif

private:
	bool unk_switch_;	///< 存在しないオプションに警告を出力するかどうか
};

}
}

#endif // #ifndef _OPTIONS_
