#ifndef _RESCORE_
#define _RESCORE_ 1
/*!
 * @brief リスコアリングを行うプログラムのヘッダ
 * @author Fujii Yasuhsia
 * @date 07/05/2009
 */

#include <vector>
#include "my_typedefs.hpp"

namespace slp {
namespace spojus {

class AcousticModel;
class AsrLangModel;
class Dictionary;
class InputData;
class Spojus;

class Rescore {
public:
	enum Mode {kQuiet, kVerbose};
	struct DpPoint {
		DpPoint() :
			score_(-INFINITY),
			x_(-1), y_(-1) {}

		void Init() {
			score_ = 0.0;
			x_ = y_ = -1;
		}

		void Set(Rnum score, int x, int y) {
			score_ = score;
			x_ = x;
			y_ = y;
		}

		Rnum score_;
		int x_;
		int y_;
	};

	Rescore(){};
	Rescore(Spojus *spojus) {
		spojus_ = spojus;
	}
	~Rescore(){
#ifdef DEBUG_DESTRUCTOR
		cout << "Rescore freed successfully" << endl;
#endif
	}

	/*!
	 * @brief 音響リスコアリングを行う
	 * @param WordContainer &words 音響リスコアリングを行う単語列
	 * @param Mode 出力のモード
	 * @param std::vector<DpPoint> *dppoint=NULL アライメント情報を残すかどうか
	 * @param bool gmm_init=false GMMの初期化を行うかどうか
	 * @retval void
	 *
	 * Null遷移の扱いが雑なので，ノードはトポロジカルソートされていること
	 *
	 */
	Rnum AmRescore(WordContainer &words, InputData *input_data, Mode mode, std::vector<DpPoint> *dppoint=NULL, bool gmm_init=false);

	/*!
	 * @brief 言語リスコアリングを行う
	 * @param WordContainer &words 言語リスコアリングを行う単語列
	 * @retval void
	 */
	Rnum LmRescore(WordContainer &words);

private:
	/*!
	 * @brief 音響リスコアリングのためのアライメントを行う
	 * @param AsrDictionary::NodeContainer& node_container アライメントするノード
	 * @param DpPoint **dp_table DPマッチングのための表
	 * @param InputData *input_data 入力データ
	 * @retval void
	 */
	void AmAlignment(AsrDictionary::NodeContainer& node_container, DpPoint **dp_table, InputData *input_data);

	/*!
	 * @brief 音響リスコアリングのためのバックトレース
	 * @param DpPoint **dp_table DPマッチングのための表
	 * @param size_t container_size ノードの数
	 * @param size_t wordsize 単語数 + 2
	 * @param int n_samples サンプル数（フレーム数）
	 * @param std::vector<DpPoint> *dppoint=NULL アライメント情報を残すかどうか
	 * @param Mode 出力のモード
	 * @retval Rnum 音響尤度
	 */
	Rnum AmBacktrace(DpPoint **dp_table, size_t container_size, size_t wordsize, int n_samples, Mode mode, std::vector<DpPoint> *dppoint);

	Spojus *spojus_;
};

}
}

#endif // #ifndef _RESCORE_
