#ifndef _SPOJUS_
#define _SPOJUS_ 1

#include <string>
#include <vector>
#include <limits>
#include <boost/thread.hpp>

#include "my_typedefs.hpp"
#include "my_templates.hpp"
#include "options.hpp"

#ifdef UTIL
bool ctrlc = false;
#endif

namespace slp {
namespace hcnf {
class HCNFWeight;
}
namespace spojus {

class AcousticModel;
class LangModel;
class AsrLangModel;
class Dictionary;
class AsrDictionary;
class TrieDictionary;
class LinearDictionary;
class OnePassDP;
class FeatureInput;
class InputData;
class MySocket;
class Lattice;
class ConfNet;
class StackDecodeLattice;

static const string kSpojusVersion = "SPOken Japanese Understanding System -SPOJUS- ver. beta";

class Spojus {
public:
	typedef vector<AcousticModel *> AcousticModelContainer;
	typedef vector<LangModel *> LangModelContainer;
	typedef vector<AsrLangModel *> AsrLangModelContainer;
	typedef vector<Dictionary *> DictionaryContainer;
	typedef vector<TrieDictionary *> TrieDictionaryContainer;
	typedef vector<LinearDictionary *> LinearDictionaryContainer;
	typedef Options::LmWeight LmWeight;
	typedef Options::LmWeightContainer LmWeightContainer;

	static const int kBuffSize = 1000;
	static const int kKaiDelay = 4;
	static const int kMaxSampleSize = 6000;

public:
	Spojus() : asr_lm_(NULL), feature_(NULL),
		   sender_(NULL), receiver_(NULL),
		   ondp_(NULL), hcnf_weight_(NULL) {};
	virtual ~Spojus(){
		Clear();
#ifdef DEBUG_DESTRUCTOR
		cout << "Spojus was freed successfully."<< endl;
#endif
	};

	const string& GetVersion() { return kSpojusVersion; }
	
	void start();
        /*!
	 * @brief 特徴量の変換を行う
	 * @param InputData *input_data 入力データ
	 * @retval void
	 */
	void FeatureConversion(InputData *input_data);

private:
	bool Init();
	void Clear();
	void ListFileInput();

	void FileInput();

	/*!
	 * @brief 入力種別がpcmファイルである場合の処理を行う
	 * @retval void
	 */
	void PcmInput();

        /*!
         * @brief mic入力をスタートさせる
         * @retval void
         */
        void MicClientStart();

        /*!
         * @brief micクライアントの終了処理
         * @retval void
         */
        void MicClientEnd();

	/*!
	 * @brief サーバとして動作させる
	 * @retval void
	 */
	void Server();

	/*!
	 * @brief Kailabフロントエンドをサーバとして動作させる
	 * @retval void
	 */
	void KaiServer();

	/*!
	 * @brief Kailabフロントエンドをサーバとして動作させる場合の認識部
	 * @retval void
	 */
	void KaiServerRecog();

	/*!
	 * @brief クライアントからの接続を待つ
	 * @retval bool
	 */
	bool ReceiveClient();

	/*!
	 * @brief フロントエンドに接続する
	 * @retval void
	 */
	void ConnectFrontend();

	/*!
	 * @brief 与えられたフレームにおけるデルタを計算する
	 * @param int frame デルタを計算するフレーム
	 * @param InputData *デルタを計算するデータ
	 * @retval void
	 */
	void ComputeDelta(int frame, InputData *input_data);

	void MakeAsrDictionary() ;

	/*!
	 * @brief 与えられた入力データについて認識を行う
	 * @param InputData *input_data 入力データ
	 * @retval void
	 */
	void Recognition(InputData *input_data);

        /*!
         * @brief 指定されたフレームの特徴量の変換を行う
         * @param int frame フレーム番号
         * @retval void
         */
        void FeatureConversion(int frame);

	/*!
	 * @brief 1pass目の認識をするための初期化処理を行う
	 * @param InputData *input_data 入力データ
	 * @retval void
	 */
	void FirstPassInit(InputData *input_data);

	/*!
	 * @brief 1パス目の探索を行う
	 * @param InputData *input_data 入力データ
	 * @retval void
	 */
	void FirstPass(InputData *input_data);

	/*!
	 * @brief 2パス目の探索を行う
	 * @param InputData *input_data 入力データ
	 * @retval Lattice * 作成したlatticeへのポインタ
	 */
	Lattice *SecondPass(InputData *input_data);

	/*!
	 * @brief Confusion Networkを用いた探索を行う
	 * @param Lattice *lattice Confusion Networkの作成に使うラティス
	 * @retval 作成されたConfusion Networkへのポインタ
	 */
	ConfNet *ConfNetSearch(Lattice *lattice);

	/*!
	 * @brief 整形処理を行う
	 * @param ConfNet *conf_net 整形処理に使うCN
	 * @retval void
	 */
	void Edit(ConfNet *conf_net);
	
	/*!
	 * @brief GMMの尤度に基づき音響モデルを選択する（辞書も選択される）
	 * @param InputData *input_data 特徴量
	 * @retval void
	 */
	void SelectAcousticModel(InputData *input_data);

	/*!
	 * @brief 言語モデルに未知語ペナルティを設定する．
	 * @retval void
	 */
	void SetUNKPenalty();

public:
	/*!
	 * @brief 音声認識用の言語モデルを作成する
	 * @retval void
	 */
	void MakeAsrLangModel();

	/*!
	 * @brief 音声認識用の言語モデルを辞書から作成する
	 * @retval void
	 */
	void MakeDicAsrLangModel();

	/*!
	 * @brief ソケットにXMLヘッダを送信する
	 *
	 */
	void SendXMLHeader(MySocket *socket);

	/*!
	 * @brief ソケットにXMLフッタを送信する
	 *
	 */
	void SendXMLFooter(MySocket *socket);

public:
	Options options_;				///< オプションクラス

	/// 複数のモデルを読み込めるようにコンテナに保存
	AcousticModelContainer am_list_;		///< 音響モデルのリスト
	LangModelContainer lm_list_;			///< 言語モデルのリスト
	AsrLangModelContainer asr_lm_list_;		///< 音声認識用言語モデルのリスト
	DictionaryContainer dic_list_;			///< 辞書のリスト
	TrieDictionaryContainer trie_dic_list_;		///< 木辞書のリスト
	LinearDictionaryContainer linear_dic_list_;	///< 線形辞書のリスト
	AcousticModelContainer gmm_list_;		///< GMMのリスト

	/// 現在使用しているモデルを保存する
	AcousticModel *am_;				///< 音響モデル
	AsrLangModel *asr_lm_;				///< 音声認識用言語モデル
	Dictionary *dic_;				///< 辞書
	TrieDictionary *trie_dic_;			///< 木辞書
	LinearDictionary *linear_dic_;			///< 線形辞書

	FeatureInput *feature_;				///< 入力データ
        FeatureInput *fi_client_;			///< クライアント用
        boost::thread *thr_fi_client_;			///< クライアントをスレッドとして起動する用

	MySocket *sender_;				///< クライアント用ソケット
	MySocket *receiver_;				///< サーバ用ソケット

	OnePassDP *ondp_;				///< 1パス目の処理を行うクラス

	hcnf::HCNFWeight *hcnf_weight_;
};


/*!
 * @brief リストに格納されたファイルからデータを読み込み，リストに格納する
 * @param list<string> &file_list ファイル名を格納したリストへの参照
 * @param string message 出力メッセージ
 * @param Container &data_list 読み込んだデータを格納するコンテナ
 * @retval void
 */
template<class Data, class Container>
void FileListRead(list<string> &file_list, string message, Container &data_list)
{
	list<string>::iterator str_iter;
	Data *model;
	for (str_iter = file_list.begin(); str_iter != file_list.end(); ++ str_iter) {
		cout << message << *str_iter << endl;
		model = new Data(*str_iter);
		data_list.push_back(model);
	}
}


/*!
 * @brief リストに格納されたファイルからデータを読み込み，リストに格納する
 * @param list<string> &file_list ファイル名を格納したリストへの参照
 * @param string message 出力メッセージ
 * @param Container &data_list 読み込んだデータを格納するコンテナ
 * @param OptionClass *opt_class コンストラクタにファイル名以外に与えるクラスへのポインタ
 * @retval void
 */
template<class Data, class Container, class OptionClass>
void FileListRead(list<string> &file_list, string message, Container &data_list, OptionClass *opt_class)
{
	list<string>::iterator str_iter;
	Data *model;
	for (str_iter = file_list.begin(); str_iter != file_list.end(); ++ str_iter) {
		cout << message << *str_iter << endl;
		model = new Data(*str_iter, opt_class);
		data_list.push_back(model);
	}
}


/*!
 * @brief ModelPackのリストに格納されたファイルからデータを読み込み，リストに格納する
 * @param const list<Pack>& file_list ModelPackを格納したリストへの参照
 * @param const string& message 出力メッセージ
 * @param Container &data_list 読み込んだデータを格納するコンテナ
 * @retval void
 */
template<class Pack, class Container>
void ModelPackListRead(list<Pack>& file_list, const string& message, Container &data_list)
{
	typename list<Pack>::iterator lp_iter;
	for (lp_iter = file_list.begin(); lp_iter != file_list.end(); ++ lp_iter) {
		cout << message << lp_iter->file_ << endl;
		lp_iter->model_->ReadModel(lp_iter->file_);
		data_list.push_back(lp_iter->model_);
	}
}


/*!
 * @brief ModelPackのリストに格納されたファイルからデータを読み込み，リストに格納する
 * @param const list<Pack>& file_list ModelPackを格納したリストへの参照
 * @param const string& message 出力メッセージ
 * @param Container &data_list 読み込んだデータを格納するコンテナ
 * @param const OptionClass *opt_class コンストラクタにファイル名以外に与えるクラスへのポインタ
 * @retval void
 */
template<class Pack, class Container, class OptionClass>
void ModelPackListRead(list<Pack>& file_list, const string& message,
		       Container &data_list, OptionClass *opt_class)
{
	typename list<Pack>::iterator lp_iter;
	for (lp_iter = file_list.begin(); lp_iter != file_list.end(); ++ lp_iter) {
		cout << message << lp_iter->file_ << endl;
		lp_iter->model_->ReadModel(lp_iter->file_, opt_class);
		data_list.push_back(lp_iter->model_);
	}
}

}
}

#endif //_SPOJUS_
