#ifndef _STACK_DECODE_LATTICE_
#define  _STACK_DECODE_LATTICE_ 1
/*!
 * @brief lattice.cppのラティスを使用して，スタックでコーディングを行うためのヘッダファイル
 * @author Fujii Yasuhisa
 * @date 03/08/2009
 */

#include <deque>

#include "lattice.hpp"
#include "my_typedefs.hpp"
#include "spojus.hpp"
#include "rescore.hpp"

namespace slp {
namespace spojus {

class StackDecodeLattice {
public:
	struct Hyp {
		Hyp() :
			heuristics_(0.0),
			score_(0.0),
			ac_score_(0.0),
			lm_score_(0.0),
			arc_(NULL) {}
		void Init() {
			heuristics_ = 0.0;
			score_ = 0.0;
			ac_score_ = 0.0;
			lm_score_ = 0.0;
			arc_ = NULL;
			words_.clear();
			pron_words_.clear();
			dppoint_.clear();
		}
		WordContainer words_;
		WordContainer pron_words_;
		Rnum heuristics_;
		Rnum score_;
		Rnum ac_score_;
		Rnum lm_score_;
		std::vector<Rescore::DpPoint> dppoint_;
		Lattice::Arc *arc_;
	};
	typedef deque<Hyp> HypContainer;

	struct HypWordsCompare {
		bool operator() (const Hyp& hyp1, const Hyp& hyp2) const { 
			size_t size = (hyp1.words_.size() < hyp2.words_.size() ?
				       hyp1.words_.size() : hyp2.words_.size());
			for (size_t i = 0; i < size; ++ i) {
				if (hyp1.words_[i] != hyp2.words_[i]) {
					return hyp1.words_[i] < hyp2.words_[i];
				}
			}
			if (hyp1.words_.size() == hyp2.words_.size()) {
				return (hyp1.score_ + hyp1.heuristics_) > (hyp2.score_ + hyp2.heuristics_);
			} else {
				return hyp1.words_.size() < hyp2.words_.size();
			}
		}
	};

	struct HypCompare {
		bool operator() (const Hyp& hyp1, const Hyp& hyp2) const { 
			return (hyp1.score_ + hyp1.heuristics_) > (hyp2.score_ + hyp2.heuristics_);
		}
	};

	struct HypEqual {
		bool operator() (const Hyp& hyp1, const Hyp& hyp2) const {
			if (hyp1.words_.size() != hyp2.words_.size()) {
				return false;
			}
			for (int i = 0; i < (int) hyp1.words_.size(); ++ i) {
				if (hyp1.words_[i] != hyp2.words_[i]) {
					return false;
				}
			}
			return true;
		}
	};

#if 0
	struct HypCompare {
		bool operator() (const Hyp& hyp1, const Hyp& hyp2) const { 
			return hyp1.heuristics_ > hyp2.heuristics_;
		}
	};
#endif

	StackDecodeLattice() : lattice_(NULL) {};
	StackDecodeLattice(Lattice *lattice, InputData *input_data, Spojus *spojus){
		lattice_ = lattice;
		input_data_ = input_data;
		spojus_ = spojus;
	};

	virtual ~StackDecodeLattice() {
#if 0
		if (lattice_ != NULL) {
			delete lattice_;
		}
		lattice_ = NULL;
#endif
	};

	void Search();
	
	/*!
	 * @brief 単語仮説を出力する
	 * @param ushort num 表示する仮説数
	 * @retval void
	 */
	void Show(ushort num);

	/*!
	 * @brief 単語仮説を出力する
	 * @param HypContainer& h_stack 仮説を出力するスタック
	 * @param ushort num 表示する仮説数
	 * @retval void
	 */
	void Show(HypContainer& h_stack, ushort num);

	/*!
	 * @brief 単語仮説をXMLで出力する
	 * @param ushort num 表示する仮説数
	 * @retval void
	 */
	void SendXML(MySocket *socket, ushort num);

	void Rescore();

	/*!
	 * @brief n + 1 番目の仮説を返す
	 * @param int n
	 * @retval Hyp *
	 */
	Hyp *GetHyp(int n) {
		return &hyp_stack_[n];
	}

	
private:
	void InitHypStack();

	/*!
	 * @brief Hypからlm_wordsにn-gramを読み込みます
	 * @param Hyp& hyp n-gramを保持した仮説
	 * @param Word *lm_words n-gramを格納する領域
	 * @retval void
	 */
	void GetNgram(Hyp& hyp, Word *lm_words);

	/*!
	 * @brief 単語履歴を束ねながらスタックに詰む
	 * @param Hyp& new_hyp スタックに詰む仮説
	 * @retval void
	 */
	void BunchPush(Hyp& new_hyp);

	/*!
	 * @brief 二つの仮説の単語履歴を調査して，次単語にとって言語モデルの履歴として同じかどうか判定する
	 * @param Hyp& hyp1 単語履歴を比較する仮説1
	 * @param Hyp& hyp2 単語履歴を比較する仮説2
	 * @retval bool true : 履歴が同じ， false : 履歴が違う
	 */
	bool CheckHistory(Hyp& hyp1, Hyp& hyp2);

	/*!
	 * @brief 仮説の展開を行う
	 * @param Hyp& hyp 展開する仮説
	 * @param Lattice::Arc *arc 接続する単語
	 * @param HypContainer& merge_stack 仮説を格納するスタック
	 * @retval void
	 */
	void HypExpansion(Hyp& hyp, Lattice::Arc *arc, HypContainer& merge_stack);

	/*!
	 * @brief スタックに積まれた仮説のソートとマージを行う
	 * @param HypContainer& hyp_stack マージ先のスタック
	 * @param HypContainer& merge_stack マージする仮説を積んだスタック
	 * @param HypContainer& tmp_stack 作業領域
	 * @retval void
	 */
	void MergeSortStack(HypContainer& hyp_stack, HypContainer& merge_stack, HypContainer& tmp_stack);

private:
	Lattice *lattice_;
	InputData *input_data_;
	Spojus *spojus_;
	HypContainer hyp_stack_;
	Word lm_words_[LangModel::kMaxNgram];
};

}
}

#endif //#ifndef _STACK_DECODE_LATTICE_
